#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 12/09/2024 11:11:20

@author: psakic
"""
import pandas as pd
import requests
import os
from datetime import datetime
from tqdm import tqdm
import re
import argparse

import autorino.common as arocmn
import autorino.download as arodwl

#### Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger('autorino')
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


def download_html_page(url_inp, output_file_inp):
    """
    Download a webpage and save its content to a file.

    Parameters
    ----------
    url_inp : str
        The URL of the webpage to download.
    output_file_inp : str
        The file path where the content will be saved.

    Returns
    -------
    str or None
        The path to the downloaded file if successful, otherwise None.
    """
    try:
        # Send a GET request to the URL
        response = requests.get(url_inp)

        # Check if the request was successful
        if response.status_code == 200:
            # Get the total file size
            total_size = int(response.headers.get("content-length", 0))
            # Write the content to the output file with a progress bar
            with open(output_file_inp, "wb") as file, tqdm(
                desc=output_file_inp,
                total=total_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as bar:
                for data in response.iter_content(chunk_size=1024):
                    file.write(data)
                    bar.update(len(data))
            logger.info(f"Page downloaded successfully and saved to {output_file_inp}")
            return output_file_inp
        else:
            logger.info(f"Failed to download page. Status code: {response.status_code}")
            return None
    except requests.RequestException as e:
        logger.info(f"An error occurred: {e}")
        return None

# Example usage


def extract_trimble_filelist(
    html_files, pattern=r".{4}______[0-9]{12}A\.T02", output_csv_dir=None
):
    """
    Extract the list of Trimble files from an HTML file.

    Parameters
    ----------
    html_files : list or str
        The path to the HTML files. If a single file is provided, it should be a string.
    pattern : str, optional
        The regex pattern to match the Trimble files. Default is ".{4}______[0-9]{12}A\.T02".
    output_csv_dir : str, optional
        The directory where the CSV files will be saved. If None, CSV files will not be saved. Default is None.

    Returns
    -------
    list
        A list of Trimble files extracted from the HTML file.
    """

    if isinstance(html_files, str):
        html_files = [html_files]

    t02_stk = []
    for html_file in html_files:
        linelis = open(html_file).readlines()

        r_stk = []

        for l in linelis:
            # DSD0______202405010000A.T02
            r = re.search(pattern, l)
            if r:
                rok = r
                r_stk.append(rok.group(0))

        if output_csv_dir and len(r_stk) > 0:
            output_csv = os.path.join(
                output_csv_dir, os.path.basename(html_file).replace(".html", ".csv")
            )
            df = pd.DataFrame(sorted(list(set(r_stk))))
            df[1] = df[0].str.extract("(2[0-9]{7})").apply(pd.to_datetime)
            df.to_csv(output_csv, index=False, header=False)
            logger.info(f"Trimble file list saved to {output_csv}")

        t02_stk = t02_stk + sorted(list(set(r_stk)))

    t02_stk = sorted(list(set(t02_stk)))
    return t02_stk

def trimble_filelist_html(
    site,
    hostname,
    output_dir,
    start_date,
    end_date,
    period="1M",
    structure="download/Internal/%Y%m",
    force=False
):

    eporng = arocmn.EpochRange(start_date, end_date, period, round_method="none")
    output_paths_ok = []
    for curr_date in eporng.eporng_list():
        url = str(os.path.join('http://', hostname, curr_date.strftime(structure)))

        #output_path_ini = os.path.join(output_dir, os.path.basename(url))
        output_fnam_ok = site + "_" + os.path.basename(url) + ".html"
        output_path_ok = str(os.path.join(output_dir, output_fnam_ok))

        if not force and os.path.exists(output_path_ok):
            logger.info(f"File {output_path_ok} already exists. Skipping download.")
            output_paths_ok.append(output_path_ok)
        else:
            logger.info(f"Downloading page from {url}")
            output_path_out = arodwl.download_http(url, output_dir)
            if output_path_out:
                os.rename(output_path_out, output_path_ok)
                output_paths_ok.append(output_path_ok)

        if output_path_ok:
            extract_trimble_filelist(output_path_ok, output_csv_dir=output_dir)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 18/09/2024 18:26:37

@author: psakic
"""



import autorino.handle as arohdl


#### Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger('autorino')
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])



def splice_rnx_rel(
    rnxs_inp,
    out_dir,
    tmp_dir,
    log_dir=None,
    handle_software="converto",
    period="1d",
    rolling_period=False,
    rolling_ref=-1,
    round_method="floor",
    drop_epoch_rnd=False,
    rinexmod_options=None,
    metadata=None,
):
    """
    Splices RINEX files together in a relative way, based on certain criteria.

    This function takes in a list of RINEX files and splices them together based on the provided criteria.
    The spliced files are stored in the specified output directory.

    Parameters
    ----------
    rnxs_inp : list
        The input RINEX files to be spliced.
        The input can be:
        * a python list
        * a text file path containing a list of files
        * a tuple containing several text files path
        * a directory path.
    out_dir : str
        The output directory where the spliced files will be stored.
    tmp_dir : str
        The temporary directory used during the splicing process.
    log_dir : str, optional
        The directory where logs will be stored. If not provided, it defaults to tmp_dir.
    handle_software : str, optional
        The software to be used for handling the RINEX files during the splice operation.
        Defaults to "converto".
    period : str, optional
        The period for splicing the RINEX files. Defaults to "1d".
    rolling_period : bool, optional
        Whether to use a rolling period for splicing the RINEX files.
        If False, the spliced files will be based only on the "full" period provided,
        i.e. Day1 00h-24h, Day2 00h-24h, etc.
        If True, the spliced files will be based on the rolling period.
        i.e. Day1 00h-Day2 00h, Day1 01h-Day2 01h, Day1 02h-Day2 02h etc.
        Defaults to False.
        see also eporng_fcts.round_epochs function
    rolling_ref :  datetime-like or int, optional
        The reference for the rolling period.
        If datetime-like object, use this epoch as reference.
        If integer, use the epoch of the corresponding index
        Use -1 for the last epoch for instance.
        The default is -1.
        see also eporng_fcts.round_epochs function
    round_method : str, optional
        The method for rounding the epochs during the splice operation. Defaults to "floor".
    drop_epoch_rnd : bool, optional
        Whether to drop the rounded epochs during the splice operation. Defaults to False.
    rinexmod_options : dict, optional
        The options for modifying the RINEX files during the splice operation. Defaults to None.
    metadata : str or list, optional
        The metadata to be included in the spliced RINEX files. Possible inputs are:
         * list of string (sitelog file paths),
         * single string (single sitelog file path)
         * single string (directory containing the sitelogs)
         * list of MetaData objects
         * single MetaData object. Defaults to None.

    Returns
    -------
    spc_main_obj : object
        The main SpliceGnss object after the splice operation.
    """
    if not log_dir:
        log_dir = tmp_dir

    spc_inp = arohdl.SpliceGnss(out_dir, tmp_dir, log_dir, metadata=metadata)
    spc_inp.load_tab_filelist(rnxs_inp)
    spc_inp.updt_epotab_rnx(use_rnx_filename_only=True)

    spc_main_obj, spc_objs_lis = spc_inp.group_by_epochs(
        period=period,
        rolling_period=rolling_period,
        rolling_ref=rolling_ref,
        round_method=round_method,
        drop_epoch_rnd=drop_epoch_rnd,
    )

    spc_main_obj.splice_core(
        handle_software=handle_software, rinexmod_options=rinexmod_options
    )

    return spc_main_obj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/01/2025 20:18:31

@author: psakic
"""

import argparse
import pandas as pd
from autorino.api.check_rnx import check_rnx


def main():
    # Get the docstring from check_rnx function for the description
    description = "Checks the presence of RINEX files in the input directory over a specified time range, and computes their completeness ratio."

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-i",
        "--input_dir_parent",
        required=True,
        help="Path to the parent input directory containing RINEX files.",
    )
    parser.add_argument(
        "-t",
        "--input_dir_structure",
        required=False,
        help="Input directory structure pattern for organizing RINEX files. (optional)",
        default="",
    )
    parser.add_argument(
        "-s",
        "--epoch_start",
        required=True,
        help="Start epoch for the time range to check."
    )
    parser.add_argument(
        "-e",
        "--epoch_end",
        required=True,
        help="End epoch for the time range to check."
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Path to the output directory for generated files (csv, plots, etc.). (optional)"
    )
    parser.add_argument(
        "-l",
        "--sites_list",
        nargs="+",
        help="List of specific sites to check. If not provided, all sites will be processed. (optional)",
        default=[],
    )

    args = parser.parse_args()

    # Call the check_rnx function
    _ = check_rnx(
        inp_dir_parent=args.input_dir_parent,
        inp_dir_structure=args.input_dir_structure,
        epoch_start=args.epoch_start,
        epoch_end=args.epoch_end,
        sites_list=args.sites_list,
        output_dir=args.output,
    )

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 20/05/2025 20:26:56

@author: psakic
"""




# Create a logger object.
import os
import time

import numpy as np
import pandas as pd
from pathlib import Path

from geodezyx import utils, conv

import autorino.common as arocmn
import autorino.convert as arocnv
import autorino.handle.handle_cls as arohdlcls

# +++ Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])

BOLD_SRT = "\033[1m"
BOLD_END = "\033[0m"


#   _____       _ _
#  / ____|     | (_)
# | (___  _ __ | |_  ___ ___
#  \___ \| '_ \| | |/ __/ _ \
#  ____) | |_) | | | (_|  __/
# |_____/| .__/|_|_|\___\___|
#        | |
#        |_|


class SpliceGnss(arohdlcls.HandleGnss):
    def __init__(
        self,
        out_dir,
        tmp_dir,
        log_dir,
        inp_dir=None,
        inp_file_regex=None,
        epoch_range=None,
        site=None,
        session=None,
        options=None,
        metadata=None,
    ):
        """
        Initialize a SpliceGnss object.

        This constructor initializes a SpliceGnss object, which is used for handling
        the splicing of RINEX files. It inherits from the HandleGnss class.

        Parameters
        ----------
        out_dir : str
            The output directory for the spliced RINEX files.
        tmp_dir : str
            The temporary directory for intermediate files.
        log_dir : str
            The directory for log files.
        inp_dir : str, optional
            The input directory for raw files. Default is None.
        inp_file_regex : str, optional
            The regular expression for filtering input files. Default is None.
        epoch_range : EpochRange, optional
            The range of epochs to be processed. Default is None.
        site : dict, optional
            Information about the site. Default is None.
        session : dict, optional
            Information about the session. Default is None.
        options : dict, optional
            Additional options for the splicing operation. Default is None.
        metadata : dict, optional
            Metadata for the splicing operation. Default is None.
        """

        super().__init__(
            out_dir=out_dir,
            tmp_dir=tmp_dir,
            log_dir=log_dir,
            inp_dir=inp_dir,
            inp_file_regex=inp_file_regex,
            epoch_range=epoch_range,
            site=site,
            session=session,
            options=options,
            metadata=metadata,
        )

    def splice(
        self,
        input_mode="find",
        input_rinexs=None,
        handle_software="converto",
        rinexmod_options=None,
        verbose=False,
        force=False,
    ):
        """
        Splice RINEX files.

        This method splices RINEX files based on the provided input. It can find local input files,
        convert a list of RINEX files to a StepGnss object, or use an existing StepGnss object.
        The splicing operation is performed using the specified software and options.

        Parameters
        ----------
        input_mode : str, optional
            The mode for finding input RINEX files. It can be:
            - "find": to find local input files.
            - "given": to use provided input RINEX files.
            Default is "find".
        input_rinexs : str or list or StepGnss, optional
            The input RINEX files. It can be:
            - A list of RINEX file paths.
            - An existing StepGnss object.
            Default is None.
        handle_software : str, optional
            The software to use for handling the RINEX files. Default is "converto".
        rinexmod_options : dict, optional
            Additional options for the RINEX modification. Default is None.
        verbose : bool, optional
            If True, prints the table for debugging purposes. Default is False.
        force : bool, optional
            If True, forces the splicing operation. Default is False.

        Returns
        -------
        None
        """
        # set the log file
        self.set_logfile()

        # Log the start of the splicing operation
        logger.info(BOLD_SRT + ">>>>>> Splicing RINEX files" + BOLD_END)

        # set ok_inp to True per default
        self.table["ok_inp"] = True

        # generate the potential local files
        self.guess_local_rnx()
        # tests if the output local files are already there
        self.check_local_files("out")
        # switch ok_inp to False if the output files are already there
        self.filter_ok_out()

        # if force is True, force the splicing operation
        if force:
            self.force("splice")

        # Find the input RINEX files
        stp_obj_rnxs_inp = self.load_input_rnxs(input_mode, input_rinexs)
        # Feed the epochs for splicing
        self.feed_by_epochs(stp_obj_rnxs_inp, mode="splice", print_table=verbose)

        # Perform the core splicing operation
        self.splice_core(
            handle_software=handle_software, rinexmod_options=rinexmod_options
        )

        # close the log file
        self.close_logfile()

        return None

    def splice_core(
        self, handle_software="converto", rinexmod_options=None, rm_inp_files=False
    ):
        """
        Perform the core splicing operation.

        This method handles the core splicing operation for RINEX files. It iterates over each row
        in the table, performs the splicing operation using the specified software, and applies
        RINEX modifications if necessary. Temporary files are removed after the operation.

        Parameters
        ----------
        handle_software : str, optional
            The software to use for handling the RINEX files. Default is "converto".
        rinexmod_options : dict, optional
            Additional options for the RINEX modification. Default is None.

        Returns
        -------
        None
        """

        self.set_tmp_dirs()

        for irow, row in self.table.iterrows():
            if not self.mono_ok_check(
                irow,
                "splice",
                fname_custom=arocmn.iso_zulu_epoch(self.table.loc[irow, "epoch_srt"]),
            ):
                continue

            logger.info(
                ">>>> Splicing %s between %s and %s",
                self.table.loc[irow, "site"],
                arocmn.iso_zulu_epoch(self.table.loc[irow, "epoch_srt"]),
                arocmn.iso_zulu_epoch(self.table.loc[irow, "epoch_end"]),
            )

            self.mono_splice(
                irow, self.tmp_dir_converted, handle_software=handle_software
            )

            if not self.table.loc[irow, "ok_out"] and self.table.loc[irow, "ok_inp"]:
                # print this only if ok_inp is True, i.e. the file should have been converted
                logger.error("unable to splice\n%s", self.table.loc[irow].to_string())
                continue

            self.mono_rinexmod(
                irow, self.tmp_dir_rinexmoded, rinexmod_options=rinexmod_options
            )

            # if rm_inp_files:
            # IMPLEMENT ME !!!!!!

            if self.tmp_dir_rinexmoded != self.out_dir:
                self.mono_mv_final(irow, self.out_dir)

        self.remov_tmp_files()
        return None

    def mono_splice(
        self, irow, out_dir=None, table_col="fpath_inp", handle_software="converto"
    ):
        """
        "on row" method

        for each row of the table, splice the 'table_col' entry,
        typically 'fpath_inp' file

        in the splice case, fpath_inp in another SpliceGnss object,
        containing the RINEXs to splice
        """

        if not self.mono_ok_check(
            irow,
            "splice (mono)",
            fname_custom=arocmn.iso_zulu_epoch(self.table.loc[irow, "epoch_srt"]),
            switch_ok_out_false=True,
        ):
            return None

        # definition of the output directory (after the action)
        if out_dir:
            out_dir_main_use = out_dir
        elif hasattr(self, "tmp_dir_converted"):
            out_dir_main_use = self.tmp_dir_converted
        else:
            out_dir_main_use = self.tmp_dir

        spc_row = self.table.loc[irow, "fpath_inp"]
        if not isinstance(spc_row, arohdlcls.HandleGnss):
            logger.error(
                f"fpath_inp is not a HandleGnss object: {self.table.loc[irow]:}"
            )
            frnx_spliced = None
        else:
            ### it is not the current object inputs which are decompressed, but the row sub object's ones
            spc_row.tmp_decmp_files, _ = spc_row.decompress()

            #### add a test here to be sure that only one epoch is inside
            out_dir_use = self.translate_path(
                out_dir_main_use, self.table.loc[irow, "epoch_srt"]
            )

            fpath_inp_lst = list(spc_row.table["fpath_inp"])

            conv_options, conv_kwoptions = self.conv_softs_opts(
                irow, handle_software=handle_software, conv_options_sup=["-cat"]
            )
            try:
                time.sleep(1)
                frnx_spliced, _ = arocnv.converter_run(
                    fpath_inp_lst,
                    out_dir_use,
                    converter=handle_software,
                    bin_options=conv_options,
                    bin_kwoptions=conv_kwoptions,
                )
            except Exception as e:
                logger.error("Error for: %s", fpath_inp_lst)
                logger.error("Exception raised: %s", e)
                frnx_spliced = None

        if frnx_spliced:
            self.table.loc[irow, "ok_out"] = True
            self.table.loc[irow, "fpath_out"] = frnx_spliced
        else:
            self.table.loc[irow, "ok_out"] = False
            # raise e

        ### it is not the current object temps which are removed, but the row sub object's ones
        spc_row.remov_tmp_files()

        return frnx_spliced





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 27 09:16:34 2023

@author: psakic
"""

import ftplib
import io

import os

# import socket
import urllib
import urllib.request
import time
from urllib.parse import urlparse
import subprocess
import re

import requests
from bs4 import BeautifulSoup
import tqdm

#### Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


# *****************************************************************************
# define Python user-defined exceptions
class AutorinoError(Exception):
    pass


class AutorinoDownloadError(AutorinoError):
    pass


class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    logger = None
    level = None
    buf = ""

    def __init__(self, logger_inp, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger_inp
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        self.logger.log(self.level, self.buf)


def join_url(protocol_inp, hostname_inp, dir_inp, fname_inp):
    """
    a wrapper to classical join to format correctly the URL
    """
    # urltambouille
    ### add the protocol if missing
    if protocol_inp and not (
        hostname_inp.startswith("http") or hostname_inp.startswith("ftp")
    ):
        prot_n_host = protocol_inp + "://" + hostname_inp
    else:
        prot_n_host = hostname_inp

    ### remove (strip) slashs in the dir path
    dirr = dir_inp.strip("/")

    ### safty warning to check a stupid copy/paste
    if fname_inp in dirr:
        logger.warning("%s 's filename also appears in dirname", fname_inp)

    url_out = os.path.join(prot_n_host, dirr, fname_inp)

    return url_out


#  ______ _______ _____
# |  ____|__   __|  __ \
# | |__     | |  | |__) |
# |  __|    | |  |  ___/
# | |       | |  | |
# |_|       |_|  |_|


def ftp_create_obj(
    hostname_inp, username=None, password=None, timeout=15, max_try=3, sleep_time=5
):
    """
    Create an FTP object, and retry in case of a timeout.

    Parameters
    ----------
    hostname_inp : str
        The hostname of the FTP server.
    username : str, optional
        The username for FTP login. Default is None.
    password : str, optional
        The password for FTP login. Default is None.
    timeout : int, optional
        The timeout for FTP connection in seconds. Default is 15 seconds.
    max_try : int, optional
        The maximum number of retry attempts in case of failure. Default is 3.
    sleep_time : int, optional
        The sleep time between retry attempts in seconds. Default is 5 seconds.

    Returns
    -------
    ftplib.FTP or None
        The FTP object if the connection is successful, otherwise None.

    Raises
    ------
    TimeoutError
        If the connection fails after the maximum number of retry attempts.
    """
    try_count = 0
    while True:
        try:
            ftp = ftplib.FTP(hostname_inp, timeout=timeout)
            if (username is not None) and (password is not None):
                ftp.login(username, password)
            return ftp
        except TimeoutError as e:
            try_count += 1
            if try_count > max_try:
                raise e
            else:
                print(e)
                time.sleep(sleep_time)
        except (OSError, ftplib.error_perm, Exception) as e:
            logger.error("Unable to create FTP object: %s", str(e))
            return None

def list_remote_ftp(
    hostname,
    remote_dir,
    username=None,
    password=None,
    timeout=15,
    max_try=3,
    ftp_obj_inp=None,
):
    """
    List files in a remote FTP directory.

    Parameters
    ----------
    hostname : str
        The hostname of the FTP server.
    remote_dir : str
        The remote directory to list files from.
    username : str, optional
        The username for FTP login. Default is None.
    password : str, optional
        The password for FTP login. Default is None.
    timeout : int, optional
        The timeout for FTP connection in seconds. Default is 15 seconds.
    max_try : int, optional
        The maximum number of retry attempts in case of failure. Default is 3.
    ftp_obj_inp : ftplib.FTP, optional
        An existing FTP object to use for the connection. Default is None.

    Returns
    -------
    list
        A list of file paths in the remote directory.

    Raises
    ------
    AutorinoDownloadError
        If the FTP connection fails or if username/password are missing.
    """

    # Clean hostname and remote directory
    # legacy hostname_use defintion (urltambouille)
    # hostname_use = hostname.replace("ftp://", "")
    # hostname_use = hostname_use.replace("/", "")

    url_parsed = urlparse(hostname)
    if url_parsed.scheme:
        hostname_use = url_parsed.netloc
    else:
        hostname_use = url_parsed.path

    # be sure there is no hostname in the remote_dir
    remote_dir_use = remote_dir.replace(hostname, "")
    remote_dir_use = remote_dir_use.replace(hostname_use, "")

    # Connect to FTP server
    if ftp_obj_inp:
        disposable_ftp_obj = False
        ftp_obj = ftp_obj_inp
    elif username is not None and password is not None:
        disposable_ftp_obj = True
        ftp_obj = ftp_create_obj(
            hostname_inp=hostname_use,
            username=username,
            password=password,
            timeout=timeout,
            max_try=max_try,
        )
    else:
        logger.error(
            "unable to create FTP object, missing username/password or input object"
        )
        raise AutorinoDownloadError

    if not ftp_obj:
        logger.error("FTP connection failed for %s", hostname_use)
        return []

    # Change to remote directory
    try:
        ftp_obj.cwd(remote_dir_use)
    except ftplib.error_perm as e:
        logger.error("FTP directory change failed: %s", str(e))
        logger.error("Wished destination: %s", remote_dir_use)
        return []
    except EOFError as e:
        logger.error("FTP cwd failed: %s %s", remote_dir_use, str(e))
        return []



    # Retrieve list of files
    try:
        file_list_bulk = ftp_obj.nlst()
    except Exception as e:
        logger.error("FTP file list failed: %s", str(e))
        file_list_bulk = []

    # current directory (.) and parent directory (..) are removed anyway
    file_list_bulk = [f for f in file_list_bulk if f not in (".", "..")]
    file_list_join = [
        join_url("", hostname_use, remote_dir_use, f.split()[-1])
        for f in file_list_bulk
    ]
    # split()[-1] is necessary for Trimble files, to have just the filename and not the size, owner, etc.

    file_list = file_list_join

    # Close connection
    if disposable_ftp_obj:
        ftp_obj.quit()

    return list(file_list)


def download_ftp(
    url,
    output_dir,
    username=None,
    password=None,
    timeout=30,
    max_try=4,
    sleep_time=5,
    ftp_obj_inp=None,
):
    """
    Download a file from an FTP server with retry logic and progress bar.

    Parameters
    ----------
    url : str
        The URL of the file to download.
    output_dir : str
        The directory where the downloaded file will be saved.
    username : str
        The username for FTP login.
        Overrided by the ftp_obj_inp parameter.
    password : str
        The password for FTP login.
        Overrided by the ftp_obj_inp parameter.
    timeout : int, optional
        The timeout for FTP connection in seconds. Default is 15 seconds.
    max_try : int, optional
        The maximum number of retry attempts in case of failure. Default is 3.
    sleep_time : int, optional
        The sleep time between retry attempts in seconds. Default is 5 seconds.
    ftp_obj_inp : ftplib.FTP, optional
        An existing FTP object to use for the connection.
        Overrides the username and password parameters.
        Default is None.

    Returns
    -------
    str
        The path to the downloaded file, or an empty string if the download failed.
        If something goes wrong, the function will raise an AutorinoDownloadError
        (no empty string or None).

    Raises
    ------
    AutorinoDownloadError
        If the download fails after the maximum number of retry attempts.
    """

    def _ftp_callback(data):
        _ftp_callback.bytes_transferred += len(data)

    urlp = urlparse(url)
    url_host = urlp.netloc
    url_dir = os.path.dirname(urlp.path)  # [1:]
    url_fname = os.path.basename(urlp.path)

    # create the FTP object
    if ftp_obj_inp:
        disposable_ftp_obj = False
        ftp_obj = ftp_obj_inp
    elif username is not None and password is not None:
        disposable_ftp_obj = True
        ftp_obj = ftp_create_obj(
            url_host,
            username=username,
            password=password,
            timeout=timeout,
            max_try=max_try,
            sleep_time=sleep_time,
        )
    else:
        logger.error(
            "unable to create FTP object, missing username/password or input object"
        )
        raise AutorinoDownloadError

    # connect to the FTP server
    if not ftp_obj:
        logger.error("FTP connection failed for %s", url)
        raise AutorinoDownloadError

    # change to the remote directory
    try:
        ftp_obj.cwd(url_dir)
    except ftplib.error_perm as e:
        logger.error("FTP directory change failed: %s", str(e))
        logger.error("Wished destination: %s", url_dir)
        raise AutorinoDownloadError

    filename = url_fname
    ftp_obj.sendcmd("TYPE I")
    file_size = ftp_obj.size(filename)
    output_path = os.path.join(output_dir, filename)
    try_count = 0

    while True:
        try:
            with tqdm.tqdm(
                total=file_size, unit="B", unit_scale=True, desc=filename
            ) as pbar, open(output_path, "wb") as f:

                _ftp_callback.bytes_transferred = 0
                ftp_obj.retrbinary(
                    "RETR " + filename,
                    lambda data: (f.write(data), pbar.update(len(data))),
                    1024,
                )
                break
        # here are all the possible exceptions that can be raised
        except Exception as e:
            try_count += 1
            if try_count > max_try:
                logger.error("download failed, max try exceeded: %s", str(e))
                raise AutorinoDownloadError
                # this error will be handled by the upper function mono_fetch
            else:
                logger.warning(
                    "download failed (%s), try %i/%i", str(e), try_count, max_try
                )
                time.sleep(sleep_time)

    f.close()

    if disposable_ftp_obj:
        ftp_obj.quit()

    return output_path


#  _    _ _______ _______ _____
# | |  | |__   __|__   __|  __ \
# | |__| |  | |     | |  | |__) |
# |  __  |  | |     | |  |  ___/
# | |  | |  | |     | |  | |
# |_|  |_|  |_|     |_|  |_|
#


def list_remote_http(hostname, remote_dir):

    url = join_url("http", hostname, remote_dir, "")

    # legacy manual join (urltambouille
    # url = os.path.join(hostname, remote_dir)
    # url = "http://" + hostname + "/" + remote_dir

    logger.debug("HTTP file list: %s", url)

    # send HTTP request
    response = requests.get(url)

    # parse HTML response
    soup = BeautifulSoup(response.content, "html.parser")

    # extract list of files
    def _parse_remote_file_http(file_list_in):
        """
        Internal function to parse the raw response of the remote file list
        """
        fillis = []
        for f in file_list_in:
            fillis.append(f.split("?")[0])
        fillis = list(sorted(set(fillis)))
        return fillis

    file_list = [
        link.get("href") for link in soup.find_all("a") if link.get("href") != "../"
    ]
    file_list = _parse_remote_file_http(file_list)
    file_list = [hostname + f for f in file_list][
        1:
    ]  # remove the 1st element, which is the folder itself

    # fsize_list = [file_size_file_http(f) for f in file_list]

    return file_list  # , fsize_list


def size_remote_file_http(url):
    req = urllib.request.Request(url, method="HEAD")
    f = urllib.request.urlopen(req)
    return f.headers["Content-Length"]


def download_http(url, output_dir, timeout=120, max_try=4, sleep_time=5):
    """
    Download a file from an HTTP server with retry logic and progress bar.

    Parameters
    ----------
    url : str
        The URL of the file to download.
    output_dir : str
        The directory where the downloaded file will be saved.
    timeout : int, optional
        The timeout for the HTTP connection in seconds. Default is 120 seconds.
    max_try : int, optional
        The maximum number of retry attempts in case of failure. Default is 4.
    sleep_time : int, optional
        The sleep time between retry attempts in seconds. Default is 5 seconds.

    Returns
    -------
    str
        The path to the downloaded file, or an empty string if the download failed.

    Raises
    ------
    AutorinoDownloadError
        If the download fails after the maximum number of retry attempts.
    """
    # Get file size
    response = requests.head(url, timeout=timeout)
    file_size = int(response.headers.get("content-length", 0))

    # Construct output path
    filename = url.split("/")[-1]
    output_path = os.path.join(output_dir, filename)

    # Download file with progress bar
    try_count = 0
    while True:
        try:
            response = requests.get(url, stream=True, timeout=timeout)
            with open(output_path, "wb") as f:
                with tqdm.tqdm(
                    total=file_size, unit="B", unit_scale=True, desc=filename
                ) as pbar:
                    for data in response.iter_content(chunk_size=1024):
                        f.write(data)
                        pbar.update(len(data))
            break
        except requests.exceptions.RequestException as e:
            try_count += 1
            if try_count > max_try:
                raise AutorinoDownloadError

            logger.warning(
                "download failed (%s), try %i/%i", str(e), try_count, max_try
            )
            time.sleep(sleep_time)

    return output_path


#  _____ _
# |  __ (_)
# | |__) | _ __   __ _
# |  ___/ | '_ \ / _` |
# | |   | | | | | (_| |
# |_|   |_|_| |_|\__, |
#                 __/ |
#                |___/


def ping(host, ping_timeout=20):
    """
    Executes the ping command and captures the output.

    This function sends a single ICMP echo request to the specified
    host and captures the round-trip time (RTT) from the output.

    Parameters
    ----------
    host : str
        The hostname or IP address to ping.
    ping_timeout : int, optional
        The timeout for the ping command in seconds. Default is 10 seconds.

    Returns
    -------
    float or None
        The round-trip time (RTT) in seconds if the ping is successful,
        otherwise None.
    """

    result = subprocess.run(
        ["ping", "-c", "1", "-W", str(ping_timeout), host],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    match = re.search(r"time=(\d+\.?\d*)\s*ms", result.stdout)

    if match:
        return float(match.group(1)) * 10**-3
    else:
        return None


def check_file_size(file_path, min_size=1000):
    """
    Check the size of a file and log a warning if it is too
    small to be useful.

    Parameters
    ----------
    file_path : str
        The path to the file to check.
    min_size : int, optional
        The minimum file size in bytes. Default is 1000 bytes.

    Returns
    -------
    ok_size : bool
        True if the file size is acceptable, otherwise False.
    file_size : int
        The size of the file in bytes

    """
    file_size = os.path.getsize(file_path)
    if file_size < min_size:
        logger.error(
            "Excluded small file (%iB < %iB): %s", file_size, min_size, file_path
        )
        ok_size = False
    else:
        ok_size = True

    return ok_size, file_size



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 20/05/2025 20:27:15

@author: psakic
"""

import autorino.handle.handle_cls as arohdlcls

# +++ Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])

BOLD_SRT = "\033[1m"
BOLD_END = "\033[0m"

class ModifyGnss(arohdlcls.HandleGnss):
    def __init__(
        self,
        out_dir,
        tmp_dir,
        log_dir,
        inp_dir=None,
        inp_file_regex=None,
        epoch_range=None,
        site=None,
        session=None,
        options=None,
        metadata=None,
    ):
        """
        Initialize a ModifyGnss object.

        This constructor initializes a ModifyGnss object,
        which is used for a stand-alone application of RinexMod
        actions on RINEX files.
        It inherits from the HandleGnss class.

        Parameters
        ----------
        out_dir : str
            The output directory for the modified RINEX files.
        tmp_dir : str
            The temporary directory for intermediate files.
        log_dir : str
            The directory for log files.
        inp_dir : str, optional
            The input directory for raw files. Default is None.
        inp_file_regex : str, optional
            The regular expression for filtering input files. Default is None.
        epoch_range : EpochRange, optional
            The range of epochs to be processed. Default is None.
        site : dict, optional
            Information about the site. Default is None.
        session : dict, optional
            Information about the session. Default is None.
        options : dict, optional
            Additional options for the modification operation. Default is None.
        metadata : dict, optional
            Metadata for the modification operation. Default is None.
        """
        super().__init__(
            out_dir=out_dir,
            tmp_dir=tmp_dir,
            log_dir=log_dir,
            inp_dir=inp_dir,
            inp_file_regex=inp_file_regex,
            epoch_range=epoch_range,
            site=site,
            session=session,
            options=options,
            metadata=metadata,
        )

    def modify(self, verbose=False, force=False, rinexmod_options=None):
        """
        Apply RINEX modifications to the data.

        This method iterates over the rows of the table and applies RINEX modifications
        using the specified options. It checks if the operation is valid for each row
        before proceeding.

        Parameters
        ----------
        verbose : bool, optional
            If True, prints the table of operations. Default is False.
        force : bool, optional
            If True, forces the rinexmod operation even if the output files already exist.
            Default is False.
        rinexmod_options : dict, optional
            The options to be used by the rinexmod function.
            If not provided, default options are used.

        Returns
        -------
        None
        """

        # set the log file
        self.set_logfile()

        # Log the start of the splitting operation
        logger.info(
            BOLD_SRT + ">>>>>> Modify RINEX files (stand-alone rinexmod)" + BOLD_END
        )

        # set the ok_inp to True per default
        self.table["ok_inp"] = True

        guess_local_rnx = False
        if guess_local_rnx:

            # special case if we downgrade the name
            if rinexmod_options and "shortname" in rinexmod_options.keys():
                shortname = rinexmod_options["shortname"]
            else:
                shortname = False

            # generate the potential local files
            self.guess_local_rnx(shortname=shortname)
            # tests if the output local files are already there
            self.check_local_files("out")
            # switch ok_inp to False if the output files are already there
            self.filter_ok_out()

        # if force is True, force the splicing operation
        if force:
            self.force("rinexmod")

        # Find the input RINEX files
        # stp_obj_rnxs_inp = self.load_input_rnxs(input_mode, input_rinexs)

        if verbose:
            self.print_table()

        for irow, row in self.table.iterrows():
            # Check if the operation is valid for the current row
            if not self.mono_ok_check(irow, "rinexmod"):
                continue

            # Update the RINEX modification options for the current row
            rinexmod_options_use = self.updt_rnxmodopts(
                rinexmod_options, irow, debug_print=False
            )

            out_dir_use = self.translate_path(
                self.out_dir, self.table.loc[irow, "epoch_srt"]
            )

            # Apply the RINEX modification using the updated options
            self.mono_rinexmod(
                irow,
                out_dir=out_dir_use,
                table_col="fpath_inp",
                rinexmod_options=rinexmod_options_use,
                check_ok_out=False
            )

        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 25 19:20:22 2024

@author: psakic
"""

import os
import re

from geodezyx import utils


##### Frontend function


def translator(path_inp, translator_dict=None, epoch_inp=None):
    """
    Frontend function to translates a given path using environment variables,
    epoch information, and a provided dictionary.

    Parameters
    ----------
    path_inp : str
        The input path to be translated.
    translator_dict : dict, optional
        A dictionary containing keywords and their corresponding values for translation. Default is None.
    epoch_inp : datetime, optional
        A datetime object containing epoch information for translation. Default is None.

    Returns
    -------
    str
        The translated path.

    Notes
    -----
    The function first translates any environment variables in the path. If epoch information is provided, it is used
    to translate any strftime aliases in the path. If a translator dictionary is provided, it is used to translate
    any keywords in the path.
    """

    path_translated = str(path_inp)
    ### translate the environnement variables
    path_translated = _translator_env_variables(path_translated)
    ### translate the epochs
    if epoch_inp:
        path_translated = _translator_epoch(path_translated, epoch_inp)
    ### translate the keywords
    if translator_dict:
        path_translated = _translator_keywords(path_translated, translator_dict)
    return path_translated


##### Internal functions


def _translator_epoch(path_inp, epoch_inp):
    """
    Translates strftime aliases in the input path using the provided epoch information.

    This function is used internally by the translator function to handle strftime aliases in the path.
    It also handles the special aliases <HOURCHAR> and <hourchar>, which are replaced with the hour of the epoch
    in uppercase and lowercase letters, respectively.

    Parameters
    ----------
    path_inp : str
        The input path containing strftime aliases to be translated.
    epoch_inp : datetime
        A datetime object containing the epoch information used for translation.

    Returns
    -------
    str
        The translated path with strftime aliases replaced with the corresponding epoch information.

    Notes
    -----
    For more information on strftime and strptime behavior, see:
    https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
    """
    path_translated = str(path_inp)
    path_translated = epoch_inp.strftime(path_translated)

    # the <HOURCHAR> and <hourchar> alias in a time information,
    # thus must be managed here
    ichar = epoch_inp.hour
    path_translated = path_translated.replace(
        "<HOURCHAR>", utils.alphabet(ichar).upper()
    )
    path_translated = path_translated.replace(
        "<hourchar>", utils.alphabet(ichar).lower()
    )

    return path_translated


def _translator_keywords(path_inp, translator_dict):
    """
    Translates keywords in the input path using the provided dictionary.

    This function is used internally by the translator function to handle keywords in the path.
    It replaces each keyword enclosed in angle brackets (e.g., <keyword>) with its corresponding value from the dictionary.
    Keywords are case-sensitive and must not be prefixed with a dollar sign ($).

    Parameters
    ----------
    path_inp : str
        The input path containing keywords to be translated.
    translator_dict : dict
        A dictionary containing keywords and their corresponding values for translation.

    Returns
    -------
    str
        The translated path with keywords replaced with their corresponding values from the dictionary.

    Notes
    -----
    If a keyword in the path does not exist in the dictionary, it will remain unchanged in the translated path.
    """
    path_translated = str(path_inp)

    # replace autorino variable (without a <$....>)
    # old regex (before 2025-01): r'<(?!.*\$).*>'
    if re.search(r"<([^$][^>]*)>", path_translated):
        for k, v in translator_dict.items():
            path_translated = path_translated.replace("<" + k + ">", str(v))

    return path_translated


def _translator_env_variables(path_inp):
    """
    Translates environment variables in the input path.

    This function is used internally by the translator function to handle environment variables in the path.
    It replaces each environment variable enclosed in angle brackets and prefixed with a dollar sign (e.g., <$VAR>)
    with its corresponding value from the system's environment variables.

    Parameters
    ----------
    path_inp : str
        The input path containing environment variables to be translated.

    Returns
    -------
    str
        The translated path with environment variables replaced with their corresponding values from the system's environment variables.

    Notes
    -----
    If an environment variable in the path does not exist in the system's environment variables, it will remain unchanged in the translated path.
    """
    path_translated = str(path_inp)

    # replace system environment variables
    if re.search(r"<\$.*>", path_translated):
        for k, v in os.environ.items():
            path_translated = path_translated.replace("<$" + k + ">", str(v))

    return path_translated


def abs_path_wrt(relative_path, base_abs_path):
    """
    Converts a relative path to an absolute path with respect to another absolute file path.

    Parameters
    ----------
    relative_path : str
        The relative path to be converted.
    base_abs_path : str
        The absolute path of the base file or directory.


    Returns
    -------
    str
        The absolute path.
    """
    if not os.path.isabs(relative_path):
        base_dir = os.path.dirname(base_abs_path)
        absolute_path = os.path.join(base_dir, relative_path)
        return os.path.abspath(absolute_path)
    else:
        return os.path.abspath(relative_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 10 15:00:40 2024

@author: psakic
"""
# Create a logger object.
import os
import time

import numpy as np
import pandas as pd
from pathlib import Path

from geodezyx import utils, conv

import autorino.common as arocmn
import autorino.convert as arocnv

# +++ Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])

BOLD_SRT = "\033[1m"
BOLD_END = "\033[0m"


class HandleGnss(arocmn.StepGnss):
    def __init__(
        self,
        out_dir,
        tmp_dir,
        log_dir,
        inp_dir=None,
        inp_file_regex=None,
        epoch_range=None,
        site=None,
        session=None,
        options=None,
        metadata=None,
    ):
        """
        Initialize a HandleGnss object.

        This constructor initializes a HandleGnss object, which is used for handling
        GNSS data processing. It inherits from the StepGnss class.

        Parameters
        ----------
        out_dir : str
            The output directory for the processed files.
        tmp_dir : str
            The temporary directory for intermediate files.
        log_dir : str
            The directory for log files.
        inp_dir : str, optional
            The input directory for raw files. Default is None.
        epoch_range : EpochRange, optional
            The range of epochs to be processed. Default is None.
        site : dict, optional
            Information about the site. Default is None.
        session : dict, optional
            Information about the session. Default is None.
        options : dict, optional
            Additional options for the processing operation. Default is None.
        metadata : dict, optional
            Metadata for the processing operation. Default is None.
        """
        super().__init__(
            out_dir=out_dir,
            tmp_dir=tmp_dir,
            log_dir=log_dir,
            inp_dir=inp_dir,
            inp_file_regex=inp_file_regex,
            epoch_range=epoch_range,
            site=site,
            session=session,
            options=options,
            metadata=metadata,
        )

    def group_by_epochs(
        self,
        period="1d",
        rolling_period=False,
        rolling_ref=-1,
        round_method="floor",
        drop_epoch_rnd=False,
    ):
        """
        Group the data by epochs.

        This method groups the data in the table by epochs, rounding the epochs according
        to the specified period and method.
        It creates a main HandleGnss object and individual
        HandleGnss objects for each epoch group.

        Parameters
        ----------
        period : str, optional
            The period for rounding the epochs. Default is "1d".
        rolling_period : bool, optional
            Whether to use a rolling period for rounding. Default is False.
        rolling_ref : int, optional
            The reference for the rolling period. Default is -1.
        round_method : str, optional
            The method for rounding the epochs. Default is "floor".
        drop_epoch_rnd : bool, optional
            Whether to drop the temporary epoch_rnd column after grouping. Default is False.

        Returns
        -------
        tuple
            A tuple containing the main HandleGnss object and a list of individual
            HandleGnss objects for each epoch group.
        """

        epoch_rnd = arocmn.round_epochs(
            self.table["epoch_srt"],
            period=period,
            rolling_period=rolling_period,
            rolling_ref=rolling_ref,
            round_method=round_method,
        )

        self.table["epoch_rnd"] = epoch_rnd

        # get the main Handle object which will describe the final spliced RINEXs
        spc_main_obj_epoch_range = arocmn.EpochRange(
            np.min(epoch_rnd),
            np.max(epoch_rnd),
            period=period,
            round_method=round_method,
        )

        spc_main_obj = SpliceGnss(
            out_dir=self.out_dir,
            tmp_dir=self.tmp_dir,
            log_dir=self.log_dir,
            epoch_range=spc_main_obj_epoch_range,
            site=self.site,
            session=self.session,
        )

        # get individual Handle objects
        grps = self.table.groupby("epoch_rnd")

        spc_obj_lis_out = []

        spc_main_obj.table["ok_inp"] = False

        logger.info("%i epoch group(s) found", len(grps))

        for i_tabgrp, (t_tabgrp, tabgrp) in enumerate(grps):
            spc_obj = self.copy()

            if drop_epoch_rnd:  # remove the temporary epoch_rnd column
                tabgrp_bis = tabgrp.drop("epoch_rnd", axis=1)
            else:  # keep the temporary epoch_rnd column
                tabgrp_bis = pd.DataFrame(tabgrp)

            logger.info(
                "epoch group #%i: from %s for %s", i_tabgrp + 1, t_tabgrp, period
            )

            spc_obj.table = tabgrp_bis
            spc_obj.updt_eporng_tab()
            spc_obj_lis_out.append(spc_obj)

            # fill the main object with the individuals
            # then "fpath_inp" is an individual SpliceGnss Object !
            spc_main_obj.table.loc[i_tabgrp, "fpath_inp"] = spc_obj
            spc_main_obj.table.loc[i_tabgrp, "fname"] = os.path.basename(
                spc_obj.table.iloc[0]["fpath_inp"]
            )
            spc_main_obj.table.loc[i_tabgrp, "ok_inp"] = True

        return spc_main_obj, spc_obj_lis_out

    def feed_by_epochs(self, step_obj_feeder, mode="split", print_table=False):
        """
        For a HandleGnss object, with a predefined epoch range
        find the corresponding RINEX for splice/split in the step_obj_feeder StepGnss,
        which a list of possible RINEXs candidates for the splice/split operation

        Parameters
        ----------
        step_obj_feeder : StepGnss
            StepGnss object, which a list of possible RINEXs candidates
            for the splice/split operation

        mode : str
            split or splice
            if split: for fpath_inp, only one RINEX is returned
            (the need one for the split)
            if splice: for fpath_inp, a SpliceGnss object with several RINEXs is returned
            (all the needed ones for the splice)

        print_table : bool, optional
            If True, prints the tables for debugging purposes. Default is False.

        Returns
        -------
        None
        """
        if not (
            self.get_step_type(full_object_name=True)
            in ("HandleGnss", "SplitGnss", "SpliceGnss")
        ):
            logger.warning(
                "feed_by_epochs works with HandleGnss, SplitGnss, SpliceGnss objects only (%s here)",
                self.get_step_type(),
            )

        if print_table:
            logger.info("> Feeding table:")
            step_obj_feeder.print_table()
            logger.info("> Table to be feeded:")
            self.print_table()

        for irow, row in self.table.iterrows():

            if not self.mono_ok_check(
                irow,
                "feed_by_epochs",
                fname_custom=arocmn.iso_zulu_epoch(self.table.loc[irow, "epoch_srt"]),
            ):
                continue

            site = self.table.loc[irow, "site"]
            epo_srt_to_feed = self.table.loc[irow, "epoch_srt"]
            epo_end_to_feed = self.table.loc[irow, "epoch_end"]

            logger.info(
                ">>>> Feeding RINEXs for %s between %s & %s",
                site,
                arocmn.iso_zulu_epoch(epo_srt_to_feed),
                arocmn.iso_zulu_epoch(epo_end_to_feed),
            )
            if mode == "splice":
                epo_srt_bol = epo_srt_to_feed <= step_obj_feeder.table["epoch_srt"]
                # For Leica, the end epoch of the RINEX can be after the theoretical one...
                # we add one hour as margin, the splice software integrates the option
                # to stop at the right epoch
                m = self.epoch_range.extra_margin_splice()
                epo_end_bol = epo_end_to_feed + m >= step_obj_feeder.table["epoch_end"]

                # epo_end_bol = np.array([True] * len(epo_end_to_feed))
            elif mode == "split":
                epo_srt_bol = step_obj_feeder.table["epoch_srt"] <= epo_srt_to_feed
                epo_end_bol = step_obj_feeder.table["epoch_end"] >= epo_end_to_feed
            else:
                logger.error("wrong mode value (accept 'splice' or 'split'): %s", mode)
                raise ValueError

            epoch_bol = epo_srt_bol & epo_end_bol

            debug = False
            if debug:
                bol_stk = pd.DataFrame(
                    np.column_stack((epo_srt_bol, epo_end_bol, epoch_bol))
                )
                print(bol_stk.to_string())

            bol_sum = np.sum(epoch_bol)

            if bol_sum == 0:
                self.table.loc[irow, "ok_inp"] = False
                self.table.loc[irow, "fpath_inp"] = None
                logger.warning(
                    "no valid input RINEX between %s & %s",
                    arocmn.iso_zulu_epoch(epo_srt_to_feed),
                    arocmn.iso_zulu_epoch(epo_end_to_feed),
                )

            elif mode == "split":
                if bol_sum > 1:
                    logger.warning("%i (>1) RINEX found for feed: %s", bol_sum)

                rnxinp_row = step_obj_feeder.table.loc[epoch_bol].iloc[0]
                ###### can be improved !
                self.table.loc[irow, "ok_inp"] = True
                self.table.loc[irow, "fpath_inp"] = rnxinp_row["fpath_inp"]
                logger.info("found for feed: %s", rnxinp_row["fpath_inp"])

            elif mode == "splice":
                spc_obj = HandleGnss(
                    out_dir=self.out_dir,
                    tmp_dir=self.tmp_dir,
                    log_dir=self.log_dir,
                    epoch_range=None,
                    site={"site_id": site},
                    session=self.session,
                )

                spc_obj.table = step_obj_feeder.table.loc[epoch_bol].copy()
                spc_obj.updt_eporng_tab()
                spc_obj.updt_site_w_rnx_fname()

                logger.info("found for feed: %s", str(spc_obj))

                self.table.loc[irow, "ok_inp"] = True
                self.table.loc[irow, "fpath_inp"] = spc_obj
            else:  # should not happend
                self.table.loc[irow, "ok_inp"] = False
                self.table.loc[irow, "fpath_inp"] = None
                logger.warning(
                    "no valid input RINEX between %s & %s",
                    arocmn.iso_zulu_epoch(epo_srt_to_feed),
                    arocmn.iso_zulu_epoch(epo_end_to_feed),
                )

        return None

    def find_local_inp(self, return_as_step_obj=True, rnx3_regex=False):
        """
        Guess the paths and name of the local raw files based on the
        EpochRange and `inp_basename` attributes of the DownloadGnss object.

        Parameters
        ----------
        return_as_step_obj : bool, optional
            If True, returns the result as a StepGnss object.
            If False, returns a list of file paths. Default is True.
        rnx3_regex : bool, optional
            If True, uses a regex pattern for RINEX 3 filenames.
            If False, uses a wildcard pattern. Default is False.

        Returns
        -------
        StepGnss or list
            If return_as_step_obj is True, returns a StepGnss object populated with the found RINEX files.
            If return_as_step_obj is False, returns a list of found file paths.
        """

        local_paths_list = []

        for epoch in self.epoch_range.eporng_list(end_bound=True):
            # guess the potential local files
            local_dir_use = self.translate_path(self.inp_dir, epoch, make_dir=False)

            if rnx3_regex:
                patrn = self.site_id9 + conv.rinex_regex_long_name()[17:]
            else:
                patrn = ".*"

            local_paths_list_epo = utils.find_recursive(
                local_dir_use, pattern=patrn, regex=True
            )
            local_paths_list.extend(local_paths_list_epo)

        logger.info("nbr local files found: %s", len(local_paths_list))
        if return_as_step_obj:
            return arocmn.rnxs2step_obj(rnxs_lis_inp=local_paths_list)
        else:
            return local_paths_list

    def load_input_rnxs(self, input_mode, input_rinexs=None):
        """
        Get the input RINEX files for handeling (splice or split).

        This method retrieves the input RINEX files based on the specified input mode and input RINEXs.
        It can find local input files or use a provided input list.

        Parameters
        ----------
        input_mode : str
            The mode for finding input RINEX files. It can be:
            - "find": to find local input files.
            - "given": to use provided input RINEX files.
        input_rinexs : str or list or StepGnss
            The input RINEX files.
            Only useful for "given" mode.
            It can be:
            - A list of RINEX file paths.
            - An existing StepGnss object.

        Returns
        -------
        StepGnss or None
            A StepGnss object containing the input RINEX files, or None if an error occurs.
        """
        method_msg = "input method to handle RINEXs: "

        if input_mode == "find":
            # Find local RINEX files and convert them to a StepGnss object
            logger.debug(
                method_msg
                + "find local RINEX files and convert them to a StepGnss object"
            )
            stp_obj_rnxs_inp = self.find_local_inp(
                return_as_step_obj=True, rnx3_regex=True
            )
        elif input_mode == "given":
            if not input_rinexs:
                logger.error("input mode is 'given' but no input_rinexs provided")

            if utils.is_iterable(input_rinexs):
                # Convert a list of RINEX files to a StepGnss object
                logger.debug(
                    method_msg + "convert a RINEX files list to a StepGnss object"
                )
                stp_obj_rnxs_inp = arocmn.rnxs2step_obj(rnxs_lis_inp=input_rinexs)

            elif isinstance(input_rinexs, arocmn.StepGnss):
                # Use an existing StepGnss object containing the RINEX files
                logger.debug(
                    method_msg + "a StepGnss object containing the RINEX files"
                )
                stp_obj_rnxs_inp = input_rinexs
            else:
                # Log an error if the input_rinexs value is invalid
                logger.error(
                    "wrong input_rinexs value for the creation of a StepGnss object: %s",
                    input_rinexs,
                )
                return None
        else:
            # Log an error if the mode value is invalid
            logger.error(
                "wrong mode value: %s (only 'find' and 'given' are valid)", input_mode
            )
            return None

        return stp_obj_rnxs_inp

    def conv_softs_opts(
        self, irow, handle_software, conv_options_sup=[], conv_kwoptions_sup=dict()
    ):
        if handle_software == "converto":
            conv_kwoptions_basis = {
                "-st": self.table.loc[irow, "epoch_srt"].strftime("%Y%m%d%H%M%S"),
                "-e": self.table.loc[irow, "epoch_end"].strftime("%Y%m%d%H%M%S"),
            }
            conv_options_basis = ["-cat"]
        elif handle_software == "gfzrnx":
            duration = int(
                (
                    self.table.loc[irow, "epoch_end"]
                    - self.table.loc[irow, "epoch_srt"]
                ).total_seconds()
            )
            conv_kwoptions_basis = {
                "-epo_beg": self.table.loc[irow, "epoch_srt"].strftime("%Y%m%d_%H%M%S"),
                "-d": duration,
            }
            conv_options_basis = ["-f"]
        else:
            logger.critical("wrong handle_software value: %s", handle_software)
            raise ValueError

        conv_option_out = conv_options_basis + conv_options_sup
        conv_kwoption_out = {**conv_kwoptions_basis, **conv_kwoptions_sup}

        return conv_option_out, conv_kwoption_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 25/10/2025 16:09:59

@author: psakic
"""

import argparse
import yaml
from autorino.api.splice_rnx_rel import splice_rnx_rel


def main():
    parser = argparse.ArgumentParser(description=splice_rnx_rel.__doc__)
    parser.add_argument(
        "-i", "--rnxs_inp",
        required=True,
        nargs="+",
        help="The input RINEX files to be spliced. The input can be a list, a text file path, a tuple of text files or a directory path."
    )
    parser.add_argument(
        "-o", "--out_dir",
        required=True,
        help="The output directory where the spliced files will be stored."
    )
    parser.add_argument(
        "-t", "--tmp_dir",
        required=True,
        help="The temporary directory used during the splicing process."
    )
    parser.add_argument(
        "-l", "--log_dir",
        default=None,
        help="The directory where logs will be stored. (optional)"
    )
    parser.add_argument(
        "-S", "--handle_software",
        default="converto",
        help="The software to be used for handling the RINEX files during the splice operation. Defaults to 'converto'. (optional)"
    )
    parser.add_argument(
        "-p", "--period",
        default="1d",
        help="The period for splicing the RINEX files. Defaults to '1d'. (optional)"
    )
    parser.add_argument(
        "-R", "--rolling_period",
        action="store_true",
        help="Whether to use a rolling period for splicing the RINEX files. (optional)"
    )
    parser.add_argument(
        "-r", "--rolling_ref",
        default="-1",
        help="The reference for the rolling period (datetime-like or int). Use -1 for the last epoch. (optional)"
    )
    parser.add_argument(
        "-m", "--round_method",
        default="floor",
        help="The method for rounding the epochs during the splice operation. Defaults to 'floor'. (optional)"
    )
    parser.add_argument(
        "-d", "--drop_epoch_rnd",
        action="store_true",
        help="Whether to drop the rounded epochs during the splice operation. (optional)"
    )
    parser.add_argument(
        "-x", "--rinexmod_options",
        type=yaml.safe_load,
        default=None,
        help="Options for modifying the RINEX files, provided as a YAML/JSON string or file content (e.g. '{longname: False}'). (optional)"
    )
    parser.add_argument(
        "-M", "--metadata",
        nargs="+",
        default=None,
        help="Metadata to include (list of sitelog paths, single sitelog path, directory, or MetaData objects). (optional)"
    )

    args = parser.parse_args()

    rolling_ref_val = args.rolling_ref

    spc_main_obj = splice_rnx_rel(
        rnxs_inp=args.rnxs_inp,
        out_dir=args.out_dir,
        tmp_dir=args.tmp_dir,
        log_dir=args.log_dir,
        handle_software=args.handle_software,
        period=args.period,
        rolling_period=args.rolling_period,
        rolling_ref=rolling_ref_val,
        round_method=args.round_method,
        drop_epoch_rnd=args.drop_epoch_rnd,
        rinexmod_options=args.rinexmod_options,
        metadata=args.metadata,
    )

    # Optionally print a short summary or representation
    if spc_main_obj is not None:
        print(spc_main_obj)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import shutil

import numpy as np
import pandas as pd

import autorino.common as arocmn
import autorino.download as arodwl
import warnings


# +++ Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])

# pd.options.mode.chained_assignment = "warn"
warnings.simplefilter("ignore", category=RuntimeWarning)

BOLD_SRT = "\033[1m"
BOLD_END = "\033[0m"


class DownloadGnss(arocmn.StepGnss):
    """
    A class to handle the downloading of GNSS data.

    Inherits from
    ----------
    arocmn.StepGnss

    Parameters
    ----------
    out_dir : str
        The output directory where the downloaded files will be stored.
    tmp_dir : str
        The temporary directory used during the download process.
    log_dir : str
        The directory where log files will be stored.
    epoch_range : EpochRange
        The range of epochs for which data will be downloaded.
    access : dict
        A dictionary containing access information such as protocol, hostname, login, and password.
    site : dict, optional
        The site information. Default is None.
    session : dict, optional
        The session information. Default is None.
    options : dict, optional
        Additional options for the download process. Default is None.
    metadata : str or list, optional
        The metadata to be included in the converted RINEX files. Possible inputs are:
         * list of string (sitelog file paths),
         * single string (single sitelog file path)
         * single string (directory containing the sitelogs)
         * list of MetaData objects
         * single MetaData object.
         Defaults to None.
    """

    def __init__(
        self,
        out_dir,
        tmp_dir,
        log_dir,
        inp_dir,
        inp_file_regex,
        epoch_range,
        access,
        site=None,
        session=None,
        options=None,
        metadata=None,
    ):
        """
        Initialize the DownloadGnss object.

        Parameters
        ----------
        out_dir : str
            The output directory where the downloaded files will be stored.
        tmp_dir : str
            The temporary directory used during the download process.
        log_dir : str
            The directory where log files will be stored.
        epoch_range : EpochRange
            The range of epochs for which data will be downloaded.
        access : dict
            A dictionary containing access information such as protocol, hostname, login, and password.
        site : dict, optional
            The site information. Default is None.
        session : dict, optional
            The session information. Default is None.
        options : dict, optional
            Additional options for the download process. Default is None.
        metadata : str or list, optional
            The metadata to be included in the converted RINEX files. Possible inputs are:
             * list of string (sitelog file paths),
             * single string (single sitelog file path)
             * single string (directory containing the sitelogs)
             * list of MetaData objects
             * single MetaData object.
             Defaults to None.
        """
        super().__init__(
            out_dir=out_dir,
            tmp_dir=tmp_dir,
            log_dir=log_dir,
            inp_dir=inp_dir,
            inp_file_regex=inp_file_regex,
            epoch_range=epoch_range,
            site=site,
            session=session,
            options=options,
            metadata=metadata,
        )

        # specific to the DownloadGnss class
        self.access = access

        # initialize the ftp object
        self.ftp_obj = None

    # legacy properties, specific to the DownloadGnss class
    #### They become more or less useless since the implementation of self.inp_file_regex (2025-01)
    @property
    def inp_dirname(self):
        return os.path.dirname(self.inp_dir)

    @property
    def inp_basename(self):
        return os.path.basename(self.inp_dir)

    def guess_remot_raw(self):
        """
        Guess the paths and name of the remote files based on the
        Session and EpochRange attributes of the DownloadGnss

        see also method ``guess_local_raw()``
        """

        if not self.inp_file_regex:
            logger.warning(
                "generic filename empty for %s, the guessed remote filepaths will be wrong",
                self.session,
            )

        hostname_use = self.access["hostname"]

        rmot_paths_list = []

        # for epoch in self.epoch_range.eporng_list():  ### go for irow ! ### IMPROVE ME !!!

        for irow, row in self.table.iterrows():
            # epoch = row["epoch_srt"]

            ### guess the potential remote files
            rmot_dir_use = str(self.inp_dir)
            rmot_fname_use = str(self.inp_file_regex)

            rmot_path_use = arodwl.join_url(
                self.access["protocol"], hostname_use, rmot_dir_use, rmot_fname_use
            )

            rmot_path_use = self.translate_path(rmot_path_use, row["epoch_srt"], make_dir=False)

            rmot_fname_use = os.path.basename(rmot_path_use)

            rmot_paths_list.append(rmot_path_use)

            # iepoch = self.table[self.table["epoch_srt"] == epoch].index[0]

            self.table.loc[irow, "fname"] = rmot_fname_use
            self.table.loc[irow, "fpath_inp"] = rmot_path_use
            logger.debug("remote file guessed: %s", rmot_path_use)

        # for guess, all input files are considered as ok a priori
        self.table["ok_inp"] = True

        rmot_paths_list = sorted(list(set(rmot_paths_list)))

        logger.info("nbr remote files guessed: %s", len(rmot_paths_list))

        return rmot_paths_list

    def guess_local_raw(self):
        """
        Guess the paths and name of the local raw files based on the
        EpochRange and `inp_file_regex` attributes of the DownloadGnss object

        see also method ``guess_remot_raw()``,
        """

        local_paths_list = []

        # for epoch in self.epoch_range.eporng_list():  # go for irow ! ### IMPROVE ME !!!

        for irow, row in self.table.iterrows():
            # epoch = row["epoch_srt"]

            # guess the potential local files
            local_dir_use = str(self.out_dir)
            local_fname_use = str(self.inp_file_regex)
            local_path_use = os.path.join(local_dir_use, local_fname_use)

            local_path_use = self.translate_path(local_path_use, row["epoch_srt"], make_dir=False)

            local_fname_use = os.path.basename(local_path_use)

            local_paths_list.append(local_path_use)

            # iepoch = self.table[self.table["epoch_srt"] == epoch].index

            self.table.loc[irow, "fname"] = local_fname_use
            self.table.loc[irow, "fpath_out"] = local_path_use
            logger.debug("local file guessed: %s", local_path_use)

        logger.info("nbr local raw files guessed: %s", len(local_paths_list))

        return local_paths_list

    def guess_remote_dirs(self):
        """
        this method is specific for ask_remote_raw
        guessing the directories is different from guessing the files:
        * no hostname
        * no filename (obviously)
        """
        rmot_dir_list = []
        for irow, row in self.table.iterrows():
            epoch = self.table.loc[irow, "epoch_srt"]
            rmot_dir_use = str(self.inp_dirname)
            rmot_dir_use = self.translate_path(rmot_dir_use, epoch, make_dir=False)
            row["fpath_inp"] = rmot_dir_use
            row["note"] = "dir_guessed"
            self.table.iloc[irow] = row

            rmot_dir_list.append(rmot_dir_use)

        rmot_dir_list = sorted(list(set(rmot_dir_list)))
        return rmot_dir_list

    def ask_remote_raw(self):
        """
        Retrieve the list of remote files from the server.

        This method guesses the remote directories and then lists the files
        in those directories based on the protocol specified in the access
        information.

        Returns
        -------
        list
            A list of remote file paths.
        """

        # guess the remote directories
        self.guess_remote_dirs()

        rmot_fil_all_lis = []
        rmot_fil_epo_lis = []
        epo_lis = []

        if self.inp_file_regex:
            logger.debug(f"regex to filter remote files (aliases not translated yet): {self.inp_file_regex}")
        else:
            logger.debug("no filter regex will be applied to remote files")

        # step 1: iterate over the input table and list the remote files (table is updated in step 2)
        for irow, row in self.table.iterrows():
            epoch = row["epoch_srt"]
            rmot_dir_use = row["fpath_inp"]

            if self.access["protocol"] == "http":
                logger.warning(
                    "HTTP protocol doesn't support well file listing. Nasty effects may occur."
                )
                rmot_fil_epo_bulk_lis = arodwl.list_remote_http(
                    self.access["hostname"], rmot_dir_use
                )
            elif self.access["protocol"] == "ftp":
                rmot_fil_epo_bulk_lis = arodwl.list_remote_ftp(
                    self.access["hostname"],
                    rmot_dir_use,
                    self.access["login"],
                    self.access["password"],
                    ftp_obj_inp=self.ftp_obj,
                )

            else:
                logger.error("wrong protocol. Only 'http' and 'ftp' are supported.")
                raise Exception

            rmot_fil_epo_bulk_lis = list(rmot_fil_epo_bulk_lis)

            ### match the right input structure, if a regex input is provided
            if self.inp_file_regex:
                rmot_fname_theo = re.compile(
                    self.translate_path(self.inp_file_regex, epoch, make_dir=False)
                )
                rmot_fil_epo_lis = [
                    f
                    for f in rmot_fil_epo_bulk_lis
                    if rmot_fname_theo.search(os.path.basename(f))
                ]
            else:
                rmot_fil_epo_lis = rmot_fil_epo_bulk_lis

            logger.debug(
                "%i remote files found on rec: %s",
                len(rmot_fil_epo_lis),
                rmot_fil_epo_lis,
            )
            if len(rmot_fil_epo_lis) != len(rmot_fil_epo_bulk_lis):
                logger.warning(
                    "%i files have been filtered out (non-matching regex)",
                    len(rmot_fil_epo_bulk_lis) - len(rmot_fil_epo_lis),
                )

            rmot_fil_all_lis = rmot_fil_all_lis + rmot_fil_epo_lis
            epo_lis = epo_lis + [epoch] * len(rmot_fil_epo_lis)

            # re add protocol: (urltambouille)
            rmot_fil_epo_lis = [
                arodwl.join_url(self.access["protocol"], "", "", f)
                for f in rmot_fil_epo_lis
            ]
            # legacy readd
            # rmot_fil_epo_lis = [
            #    self.access["protocol"] + "://" + f for f in rmot_fil_epo_lis
            # ]

            ## step 2: the table is updated with the files found
            new_rows_stk = []
            for rmot_fil in rmot_fil_epo_lis:
                new_row = row.copy()
                new_row["fname"] = os.path.basename(rmot_fil)
                new_row["fpath_inp"] = rmot_fil
                new_row["note"] = ""
                new_row["ok_inp"] = True
                new_rows_stk.append(new_row)

            self.table = pd.concat(
                [self.table, pd.DataFrame(new_rows_stk)], ignore_index=True
            )

        # step 3: remove the guessed directories
        self.table = self.table[self.table["note"] != "dir_guessed"]
        self.table.reset_index(drop=True, inplace=True)

        logger.info("nbr remote files found on rec: %s", len(rmot_fil_all_lis))
        return rmot_fil_all_lis

    def ask_local_raw(self):
        """
        Guess the paths and name of the local raw files based on the
        EpochRange and  table's fpath_inp basename i.e. the theoretical remote file name
        generated with `guess_remot_raw()`.

        see also method ``guess_remot_raw()``,
        """

        local_paths_list = []

        for irow, row in self.table.iterrows():
            epoch = row["epoch_srt"]

            # guess the potential local files
            local_dir_use = str(self.out_dir)
            local_fname_use = os.path.basename(row["fpath_inp"])
            local_path_use = os.path.join(local_dir_use, local_fname_use)
            local_path_use = self.translate_path(local_path_use, epoch, make_dir=False)

            local_paths_list.append(local_path_use)

            self.table.loc[irow, "fname"] = local_fname_use
            self.table.loc[irow, "fpath_out"] = local_path_use
            logger.debug("local file asked: %s", local_path_use)

        logger.info("nbr local raw files asked: %s", len(local_paths_list))

        return local_paths_list

    def ping_remote(self, ping_max_try=4, ping_timeout=20):
        """
        Ping the remote server to check if it is reachable.
        """
        count = 0
        ping_out = None
        while count < ping_max_try and not ping_out:
            ping_out = arodwl.ping(
                host=self.access["hostname"], ping_timeout=ping_timeout
            )
            count += 1
            if count > 1:
                logger.warning(
                    "attempt %i/%i to ping %s",
                    count,
                    ping_max_try,
                    self.access["hostname"],
                )

        if not ping_out:
            logger.error("Remote server %s is not reachable.", self.access["hostname"])
            self.exit_code = 7
        else:
            logger.info(
                "Remote server %s is reachable. (%d ms)",
                self.access["hostname"],
                int(ping_out * 10**3),
            )

        return ping_out

    def set_ftp_obj(self, timeout=15, max_try=4, sleep_time=5):
        """
        Create the FTP object for the FTP protocol.
        """
        self.ftp_obj = arodwl.ftp_create_obj(
            hostname_inp=self.access["hostname"],
            username=self.access["login"],
            password=self.access["password"],
            timeout=timeout,
            max_try=max_try,
            sleep_time=sleep_time,
        )

    def download(
        self,
        verbose=False,
        force=False,
        remote_find_method="ask",
        invalidate_small_local_files=True,
        timeout=60,
        max_try=4,
        sleep_time=5,
        ping_max_try=4,
        ping_timeout=20,
        ping_disable=False,
    ):
        """
        Frontend method to download files from a GNSS receiver

        Parameters
        ----------
        verbose : bool, optional
            If True, prints detailed information during the download process. Default is False.
        force : bool, optional
            If True, forces the download even if the files already exist locally. Default is False.
        remote_find_method : str, optional
            Method to find remote files.
            Can be 'ask' to list files from the server (for FTP only) or 'guess' to guess file paths.
            Default is 'ask'.
        invalidate_small_local_files : bool, optional
            If True, invalidates small local files to ensure they are re-downloaded. Default is True.
        timeout : int, optional
            Timeout in seconds for the download operations. Default is 60.
        max_try : int, optional
            Maximum number of retry attempts for the download operations. Default is 4.
        sleep_time : int, optional
            Sleep time in seconds between retry attempts. Default is 5.
        ping_max_try : int, optional
            Maximum number of retry attempts for pinging the remote server. Default is 4.
        ping_timeout : int, optional
            Timeout in seconds for pinging the remote server. Default is 20.
        ping_disable : bool, optional
            If True, skips the pinging of the remote server. Default is False.

        Returns
        -------
        None
        """
        self.set_logfile()
        logger.info(BOLD_SRT + ">>>>>> RAW files download" + BOLD_END)

        # Set up and clean temporary directories
        self.set_tmp_dirs()
        self.clean_tmp_dirs()

        # Check the remote find method, and switch to 'guess' if HTTP protocol is used
        if remote_find_method == "ask" and self.access["protocol"] == "http":
            logger.warning("HTTP protocol doesn't support file listing ('ask' method).")
            logger.warning("Switching to 'guess' remote find method.")
            remote_find_method = "guess"

        # Ping the remote server to check if it is reachable (unless ping_disable is True)
        ping_out = True if ping_disable else self.ping_remote(ping_max_try, ping_timeout)

        if not ping_out:
            # local raw are guessed anyway, to resume the next steps if download is not possible
            self.guess_local_raw()
            return None

        # Set up the DownloadGnss's FTP object if the protocol is FTP
        if self.access["protocol"] == "ftp":
            self.set_ftp_obj(timeout=timeout, max_try=max_try, sleep_time=sleep_time)

        # Guess remote raw file paths
        if remote_find_method == "guess":
            self.guess_remot_raw()
            self.guess_local_raw()
        # Ask remote raw file paths (works for FTP only!
        elif remote_find_method == "ask":
            self.ask_remote_raw()
            self.ask_local_raw()
        else:
            logger.error(
                "Wrong remote_find_method: %s ('ask' or 'guess' only are allowed)",
                remote_find_method,
            )
            raise Exception

        # Check local files and update table
        self.check_local_files()
        # be sure ok_xxx columns are booleans
        self.table_ok_cols_bool()
        if invalidate_small_local_files:
            self.invalidate_small_local_files()
        # switch ok_inp to False if the output files are already there
        self.filter_ok_out()

        # Force download if required
        if force:
            self.force(step_name="download")

        # Log the number of files to be downloaded and excluded
        n_ok_inp = (self.table["ok_inp"]).sum()
        n_not_ok_inp = np.logical_not(self.table["ok_inp"]).sum()
        n_tot_inp = len(self.table)

        logger.info(
            "%3i/%3i files will be downloaded, %3i/%3i files are excluded",
            n_ok_inp,
            n_tot_inp,
            n_not_ok_inp,
            n_tot_inp,
        )

        # Print the table if verbose is enabled
        if verbose:
            self.print_table()

        # Create a lockfile to ensure exclusive access during download
        lock = self.create_lockfile()

        ###############################
        # +++ DOWNLOAD CORE a.k.a FETCH
        lock.acquire()
        try:
            self.fetch_remote_files(
                force=force,  # force argument is now redudant, because ok_inp can be forced with .force() method
                timeout=timeout,
                max_try=max_try,
                sleep_time=sleep_time,
            )
        finally:
            lock.release()
            os.remove(lock.lock_file)
        ###############################

        # Print the table if verbose is enabled
        if verbose:
            self.print_table()

        # close the log file
        self.close_logfile()

        return None

    def fetch_remote_files(self, force=False, timeout=60, max_try=4, sleep_time=5):
        """
        will download locally the files which have been identified by
        the guess_remote_files method

        exploits the fname_remote column of the DownloadGnss.table
        attribute

        This `fetch_remote_files` method is for the download stricly speaking.
        Ìn operation, use the `download` method which does a broader
        preliminary actions.
        """
        download_files_list = []

        for irow, row in self.table.iterrows():
            file_dl_out = self.mono_fetch(
                irow,
                force=force,  # force argument is now redudant, because ok_inp can be forced with .force() method
                timeout=timeout,
                max_try=max_try,
                sleep_time=sleep_time,
            )
            if file_dl_out:
                download_files_list.append(file_dl_out)

        return download_files_list

    #               _   _                   _ _                           _ _    __                                 __
    #     /\       | | (_)                 ( | )                         ( | )  / /                                 \ \
    #    /  \   ___| |_ _  ___  _ __  ___   V V_ __ ___   ___  _ __   ___ V V  | | ___  _ __    _ __ _____      _____| |
    #   / /\ \ / __| __| |/ _ \| '_ \/ __|    | '_ ` _ \ / _ \| '_ \ / _ \     | |/ _ \| '_ \  | '__/ _ \ \ /\ / / __| |
    #  / ____ \ (__| |_| | (_) | | | \__ \    | | | | | | (_) | | | | (_) |    | | (_) | | | | | | | (_) \ V  V /\__ \ |
    # /_/    \_\___|\__|_|\___/|_| |_|___/    |_| |_| |_|\___/|_| |_|\___/     | |\___/|_| |_| |_|  \___/ \_/\_/ |___/ |
    #                                                                           \_\                                 /_/

    def mono_fetch(self, irow, force=False, timeout=60, max_try=4, sleep_time=5):

        if not self.mono_ok_check(irow, "fetch"):
            return None

        logger.info(">>>> fetch remote raw file: %s", self.table.loc[irow, "fname"])

        # ++++++ use the guessed local file as destination or the generic directory
        if not arocmn.is_ok(self.table.loc[irow, "fpath_out"]):
            # +++ the local file has not been guessed
            outdir_use = str(self.out_dir)
            outdir_use = self.translate_path(
                outdir_use, self.table.loc[irow, "epoch_srt"], make_dir=True
            )
        else:  # +++ the local file has been guessed before
            outdir_use = os.path.dirname(self.table.loc[irow, "fpath_out"])

        tmpdir_use = self.tmp_dir_downloaded

        # +++++ create the directory if it does not exist
        if not os.path.exists(outdir_use):
            os.makedirs(outdir_use)

        # +++++ download the file
        file_dl_tmp = None
        file_dl_out = None
        if not self.access["protocol"] in ("ftp", "http"):
            logger.critical("wrong protocol %s", self.access["protocol"])
            raise Exception
        elif self.access["protocol"] == "http":
            try:
                file_dl_tmp = arodwl.download_http(
                    url=self.table.loc[irow, "fpath_inp"],
                    output_dir=tmpdir_use,
                    timeout=timeout,
                    max_try=max_try,
                    sleep_time=sleep_time,
                )
                dl_ok = True
            except Exception as e:
                logger.error("HTTP download error: %s", str(e))
                dl_ok = False

        elif self.access["protocol"] == "ftp":
            try:
                file_dl_tmp = arodwl.download_ftp(
                    self.table.loc[irow, "fpath_inp"],
                    tmpdir_use,
                    username=self.access["login"],
                    password=self.access["password"],
                    timeout=timeout,
                    max_try=max_try,
                    sleep_time=sleep_time,
                    ftp_obj_inp=self.ftp_obj,
                )
                dl_ok = True
            except Exception as e:
                logger.error("FTP download error: %s", str(e))
                dl_ok = False

        else:  # ++ this case should never happen since there is a protocol test at the begining
            dl_ok = False
            pass

        # +++++ check the downloaded file size
        if dl_ok:
            dl_ok, _ = arodwl.check_file_size(file_dl_tmp)

        # +++++ store the results in the table
        if dl_ok:
            file_dl_out = shutil.copy(file_dl_tmp, outdir_use)
            self.table.loc[irow, "ok_out"] = True
            self.table.loc[irow, "fpath_out"] = file_dl_out
            os.remove(file_dl_tmp)
        else:
            self.table.loc[irow, "ok_out"] = False

        return file_dl_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  8 15:47:58 2024

@author: psakic

This module, eporng_cls.py, provides a class for handling ranges of epochs.
"""

import re

import numpy as np
import pandas as pd
from geodezyx import utils

import autorino.common as arocmn

#### Import the logger
import logging
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


class EpochRange:
    """
    A class used to represent a range of epochs.

    ...

    Attributes
    ----------
    period : str
        the rounding period. Use the pandas' frequency aliases convention.
    round_method : str
        the method used for rounding the epochs.
    tz : str
        the timezone used for the epochs.
    _epoch1_raw : datetime
        the raw start of the epoch range.
    _epoch2_raw : datetime
        the raw end of the epoch range.

    Methods
    -------
    eporng_list(end_bound=False):
        Computes the list of epochs corresponding to the EpochRange.
    is_valid():
        Checks if the epoch range is valid.
    """

    def __init__(
        self, epoch1, epoch2=None, period="1d", round_method="floor", tz="UTC"
    ):
        """
        Constructs all the necessary attributes for the epoch range object.

        Parameters
        ----------
            epoch1 : str, datetime, pd.Timestamp, pd.NaT, list
                the start of the epoch range.
            epoch2 : str, datetime, pd.Timestamp, pd.NaT
                the end of the epoch range.
            period : str, optional
                the rounding period. Use the pandas' frequency aliases convention.
            round_method : str, optional
                the method used for rounding the epochs.
            tz : str, optional
                the timezone used for the epochs.
        """

        self._epoch1_raw = epoch1
        self._epoch2_raw = epoch2

        self.period = period
        self.round_method = round_method
        self.tz = tz

        if (
            self._epoch1_raw and self._epoch2_raw
        ):  # 1) regular case: a start and an end are given
            _epoch1tmp = arocmn.datepars_intrpt(self._epoch1_raw)
            _epoch2tmp = arocmn.datepars_intrpt(self._epoch2_raw)
            _epoch_min_tmp = np.min((_epoch1tmp, _epoch2tmp))
            _epoch_max_tmp = np.max((_epoch1tmp, _epoch2tmp))

            self.epoch_start = _epoch_min_tmp  ### setter bellow
            self.epoch_end = _epoch_max_tmp  ### setter bellow

            self.manual_range = False
            self._manu_range_list = []

        elif (
            utils.is_iterable(self._epoch1_raw) and not self._epoch2_raw
        ):  # 2) case a start is given as a list, but no end
            _epoch1tmp = [arocmn.datepars_intrpt(e) for e in self._epoch1_raw]
            _epoch_min_tmp = np.min(_epoch1tmp)
            _epoch_max_tmp = np.max(_epoch1tmp)

            self.epoch_start = _epoch_min_tmp
            self.epoch_end = _epoch_max_tmp

            self.manual_range = True
            self._manu_range_list = _epoch1tmp

    ## NB: I think it is a bad idea to have an attribute (property) to get the list of epochs

    def __repr__(self):
        return "from {} to {}, period {}".format(
            arocmn.iso_zulu_epoch(self.epoch_start),
            arocmn.iso_zulu_epoch(self.epoch_end),
            self.period,
        )

    ############ getters and setters
    @property
    def epoch_start(self):
        """Gets the start of the epoch range."""
        return self._epoch_start

    @epoch_start.setter
    def epoch_start(self, value):
        """Sets the start of the epoch range."""
        self._epoch_start = arocmn.datepars_intrpt(value, tz=self.tz)
        self._epoch_start = arocmn.round_date(
            self._epoch_start, self.period, self.round_method
        )

    @property
    def epoch_end(self):
        """Gets the end of the epoch range."""
        return self._epoch_end

    @epoch_end.setter
    def epoch_end(self, value):
        """Sets the end of the epoch range."""
        self._epoch_end = arocmn.datepars_intrpt(value, tz=self.tz)
        self._epoch_end = arocmn.round_date(
            self._epoch_end, self.period, self.round_method
        )

    @property
    def period_values(self):
        """
        For a period, e.g. 15min, 1H...
        Returns the value (e.g. 15, 1) and the unit (e.g. min, H)
        """
        numbers = re.findall(r"[0-9]+", self.period)
        alphabets = re.findall(r"[a-zA-Z]+", self.period)
        val = int("".join(*numbers))
        unit = str("".join(*alphabets))
        return val, unit

    @property
    def period_as_timedelta(self):
        """
        For a period, e.g. 15min, 1H...
        return in as a pandas Timedelta
        """
        return pd.Timedelta(self.period)

    ########### methods
    def eporng_list(self, end_bound=False):
        """
        Compute the list of epochs corresponding to the EpochRange.

        Parameters
        ----------
        end_bound : bool, optional
            If True, gives the end bound of the range. Default is False.

        Returns
        -------
        list
            List of epochs.
        """
        if self.manual_range:
            return self.eporng_list_manual(end_bound=end_bound)
        else:
            return self.eporng_list_regular(end_bound=end_bound)

    def eporng_list_manual(self, end_bound=False):
        """
        Compute the list of epochs for a forced range.

        Parameters
        ----------
        end_bound : bool, optional
            If True, gives the end bound of the range. Default is False.

        Returns
        -------
        list
            List of epochs.
        """
        if not self.eporng_list_manual:
            logger.error("No forced range list available")
            return []

        if not end_bound:
            return self._manu_range_list
        else:
            # subtract also one second for security reason
            plus_one = pd.Timedelta(self.period)
            return list(np.array(self._manu_range_list) + plus_one - pd.Timedelta("1s"))

    def eporng_list_regular(self, end_bound=False):
        """
        Compute the list of epochs corresponding to the EpochRange
        if end_bound = True, give the end bound of the range
        (start bound is generated per default)

        Parameters
        ----------
        end_bound : bool, optional
            If True, gives the end bound of the range.

        Returns
        -------
        list
            List of epochs.
        """
        if not self.is_valid():  ### NaT case
            eporng = [pd.NaT]
        elif not end_bound:  ### start bound
            eprrng_srt = pd.date_range(
                self.epoch_start, self.epoch_end, freq=self.period
            )
            eporng = eprrng_srt
        else:  ### end bound
            plus_one = self.period_as_timedelta
            eprrng_end = pd.date_range(
                self.epoch_start, self.epoch_end + plus_one, freq=self.period
            )
            # subtract also one second for security reason
            # first element is the epoch start, thus we remove it
            eporng = eprrng_end[1:] - np.timedelta64(1, "s")

        return list(eporng)

    def is_valid(self):
        """
        Checks if the epoch range is valid.

        Returns
        -------
        bool
            True if the epoch range is valid, False otherwise.
        """
        if pd.isna(self.epoch_start) or pd.isna(self.epoch_end):
            return False
        else:
            return True


    def extra_margin_splice(self):
        """
        Returns the extra margin for splicing operations.

        Leica raw files  can be a bit over their nominal end,
        so we need to add a margin to the splicing operation.

        Returns
        -------

        """
        if self.period_as_timedelta >= pd.Timedelta("1 day"):
            return pd.Timedelta("1 hour")
        else:
            return pd.Timedelta("1 minute")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  8 16:53:51 2024

@author: psakic
"""

import os
import re
import shutil

import numpy as np
import pandas as pd

from geodezyx import utils, conv
from filelock import FileLock, Timeout

#### Import the logger
import logging
import autorino.cfgenv.env_read as aroenv
import autorino.common as arocmn

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


def make_site_id9(site_id_inp):
    """
    Converts a site ID to a 9-character format.

    This function takes a site ID and converts it to a 9-character format.
    If the input site ID is already 9 characters long, it returns the uppercase version of the input.
    If the input site ID is 4 characters long, it appends '00XXX' to the uppercase version of the input.
    Otherwise, it takes the first 4 characters of the input, converts them to uppercase, and appends '00XXX'.

    Parameters
    ----------
    site_id_inp : str
        The input site ID to be converted.

    Returns
    -------
    str
        The site ID in 9-character format.
    """
    if len(site_id_inp) == 9:
        return site_id_inp.upper()
    elif len(site_id_inp) == 4:
        return site_id_inp.upper() + "00XXX"
    else:
        return site_id_inp[:4].upper() + "00XXX"

def dummy_site_dic():
    """
    Creates a dummy site dictionary.

    This function creates and returns a dictionary with dummy site information.
    The dictionary contains the following keys:
    - name: The name of the site.
    - site_id: The site identifier.
    - domes: The DOMES number of the site.
    - sitelog_path: The path to the site log.
    - position_xyz: The XYZ coordinates of the site.

    Returns
    -------
    dict
        A dictionary containing dummy site information.
    """
    d = dict()

    d["name"] = "XXXX"
    d["site_id"] = "XXXX00XXX"
    d["domes"] = "00000X000"
    d["sitelog_path"] = "/null"
    d["position_xyz"] = (6378000, 0, 0)

    return d


def dummy_sess_dic():
    """
    Creates a dummy session dictionary.

    This function creates and returns a dictionary with dummy session information.
    The dictionary contains the following keys:
    - name: The name of the session.
    - data_frequency: The data frequency of the session.
    - tmp_dir_parent: The parent directory for temporary files.
    - tmp_dir_structure: The directory structure for temporary files.
    - log_parent_dir: The parent directory for log files.
    - log_dir_structure: The directory structure for log files.
    - out_dir_parent: The parent directory for output files.
    - out_structure: The directory structure for output files.

    Returns
    -------
    dict
        A dictionary containing dummy session information.
    """
    d = dict()

    d["name"] = "NA"
    d["data_frequency"] = "30S"
    d["tmp_dir_parent"] = "/tmp/"
    d["tmp_dir_structure"] = "<site_id9>/%Y/%j"
    d["log_parent_dir"] = "/tmp/"
    d["log_dir_structure"] = "<site_id9>/%Y/%j"
    d["out_dir_parent"] = "/tmp/"
    d["out_structure"] = "<site_id9>/%Y/%j"

    return d


def import_files(inp_fil, inp_regex=".*"):
    """
    Handles multiple types of input lists and returns a python list of the input.

    This function can handle various types of input lists and convert them into a python list.
    The input can be:
     * a python list
     * a text file path containing a list of files
     * a tuple containing several text files path
     * a directory path.
    If the input is a directory path, all the files matching the regular expression specified by 'inp_regex' are read.

    Parameters
    ----------
    inp_fil : list or str or tuple
        The input list to be interpreted. It can be a python list, a text file path containing a list of files,
        a tuple containing several text files path, or a directory path.
    inp_regex : str, optional
        The regular expression used to filter the files when 'inp_fil' is a directory path.
        Default is ".*" which matches any file.

    Returns
    -------
    list
        The interpreted list.
    """

    if not inp_fil:
        flist = []
    # the input is a tuple, i.e. a list of text files, the output is the concatenanted content of the text files
    elif isinstance(inp_fil, tuple) and os.path.isfile(inp_fil[0]):
        flist = list(np.hstack([open(f, "r+").readlines() for f in inp_fil]))
        flist = [f.strip() for f in flist]
    # the input is a python list, the output is the same python list
    elif isinstance(inp_fil, list):
        flist = inp_fil
    # the input is a single text file path, the output is the content of the text file
    elif os.path.isfile(inp_fil):
        flist = open(inp_fil, "r+").readlines()
        flist = [f.strip() for f in flist]
    # The input is a directory path, the output is the list of files inside the directory
    elif os.path.isdir(inp_fil):
        # Here we find everything ".*", the regex will be filtered bellow
        flist = utils.find_recursive(inp_fil, ".*", regex=True)
    else:
        flist = []
        logger.warning("the filelist is empty")

    if inp_regex != ".*":
        flist = [f for f in flist if re.match(inp_regex, os.path.basename(f))]
        # os.path.basename is used to match the regex on the filename only

    return flist


def load_previous_tables(log_dir):
    """
    Load all previous tables from the log directory.

    This function searches for all files ending with "*table.log" in the log directory
    and concatenates them into a single pandas DataFrame. If no such files are found,
    it returns an empty DataFrame and logs a warning.

    Parameters
    ----------
    log_dir : str
        The directory where the log files are stored.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the concatenated data from all found log files. If no files are found,
        an empty DataFrame is returned.
    """
    # Find all files ending with "*table.log" in the log directory
    tables_files = utils.find_recursive(log_dir, "*_table.log")

    # If no such files are found, log a warning and return an empty DataFrame
    if not tables_files:
        logger.warning("No previous tables found in the log directory.")
        return pd.DataFrame([])

    # Read and concatenate non-empty DataFrames from the found files
    tab_df_stk = [pd.read_csv(t) for t in tables_files if not pd.read_csv(t).empty]

    return pd.concat(tab_df_stk, ignore_index=True) if tab_df_stk else pd.DataFrame([])

def print_tab_core(table_inp, max_colwidth=33):
    """
    Prints the table of the StepGnss object with specified formatting.

    This method formats and prints the table of the StepGnss object. It shrinks the strings in the 'fraw',
    'fpath_inp', and 'fpath_out' columns to a specified maximum length and formats the 'epoch_srt' and 'epoch_end'
    columns as strings
    with a specific date-time format. The method then prints the formatted table to the logger.

    Parameters
    ----------
    table_inp : pd.DataFrame
        The input table to be formatted and printed.

    max_colwidth : int, optional
        The maximum column width for the output table. Default is 33.

    Returns
    -------
    str
        The formatted table as a string.
    """

    def _shrink_str(str_inp, maxlen=max_colwidth):
        """
        Shrinks a string to a specified maximum length.

        This function shrinks a string to a specified maximum length by keeping the first and last parts
        of the string and replacing the middle part with '..'.
        The length of the first and last parts is half of the maximum length.

        Parameters
        ----------
        str_inp : str
            The input string to be shrunk.
        maxlen : int, optional
            The maximum length of the output string.
            Default is the value of the 'max_colwidth' parameter of the
            'verbose' method.

        Returns
        -------
        str
            The shrunk string.
        """
        if len(str_inp) <= maxlen:
            return str_inp
        else:
            halflen = int((maxlen / 2) - 1)
            str_out_shrink = str_inp[:halflen] + ".." + str_inp[-halflen:]
            return str_out_shrink

    # we define the FORMATTERS (i.e. functions) for each column
    form = dict()
    form["fraw"] = _shrink_str
    form["fpath_inp"] = _shrink_str
    form["fpath_out"] = _shrink_str

    print_time = lambda t: (
        t.strftime("%y-%m-%d %H:%M:%S") if not pd.isna(t) else "NaT"
    )
    form["epoch_srt"] = print_time
    form["epoch_end"] = print_time

    str_out = table_inp.to_string(max_colwidth=max_colwidth + 1, formatters=form)
    # add +1 in max_colwidth for safety

    return str_out


def is_ok(val_inp):
    """
    Checks if the input value is OK or not.

    This function takes an input value and checks if it is defined.
    It considers None, NaN, False and an empty string as not OK.
    If the input value is one of these, the function returns False. Otherwise, it returns True.

    Parameters
    ----------
    val_inp : any or iterable of any
        The input value to be checked.
        Can handle iterables.

    Returns
    -------
    bool
        True if the input value is defined, False otherwise.
    """

    # iterable case
    typ = utils.get_type_smart(val_inp)

    if utils.is_iterable(val_inp):
        return typ([is_ok(v) for v in val_inp])

    # scalar case
    if val_inp is None:
        return False
    elif isinstance(val_inp, float) and np.isnan(val_inp):
        return False
    elif val_inp == "":
        return False
    elif not val_inp:  # == False
        return False
    else:
        return True


def check_lockfile(lockfile_path):
    """
    Checks the lock status of a specified file.

    This function attempts to acquire a lock on the specified file. If the lock is acquired,
    it prints a success message and releases the lock. If the lock is not acquired (i.e., the file
    is already locked by a previous process), it prints a message indicating that the process is locked.

    Parameters
    ----------
    lockfile_path : str
        The path of the lock file to check the lock status for.

    Returns
    -------
    None
    """
    lock = FileLock(lockfile_path)
    try:
        lock.acquire(timeout=0)
        logger.debug(f"Lock free for {lockfile_path}")
        lock.release()
    except Timeout:
        logger.warning(f"Process is locked by a previous process for {lockfile_path}")


def rnxs2step_obj(rnxs_lis_inp):
    """
    Convert a list of RINEX files to a StepGnss object.

    This function creates a StepGnss object and populates it with data from the provided list of RINEX files.
    It loads the table from the file list, updates the site information from the RINEX filenames, and updates
    the epoch table from the RINEX filenames.

    Parameters
    ----------
    rnxs_lis_inp : list
        A list of RINEX file paths to be converted into a StepGnss object.

    Returns
    -------
    StepGnss
        A StepGnss object populated with data from the provided RINEX files.
    """
    stp_obj = arocmn.StepGnss(out_dir="", tmp_dir="", log_dir="", inp_dir="")

    stp_obj.load_tab_filelist(rnxs_lis_inp)
    stp_obj.updt_site_w_rnx_fname()
    stp_obj.updt_epotab_rnx(use_rnx_filename_only=True, update_epoch_range=True)

    return stp_obj


def guess_sites_list(inp_fil):
    """
    Guess the list of sites from the input list.

    This function guesses the list of sites from the input list.
    
    It extracts the site identifier from the input list
    and returns a list of unique site identifiers.

    Parameters
    ----------
    inp_fil : list
        The input list to extract the site identifiers from.
        See `import_files` for details.
        
    Returns
    -------
    list
        A list of unique site identifiers extracted from the input list.
    """
    inp_list = import_files(inp_fil)

    sites_list = []

    for f in inp_list:
        if conv.rinex_regex_search_tester(f, short_name=False):
            site = os.path.basename(f)[:9]
            sites_list.append(site)

    sites_list = list(sorted(list(set(sites_list))))

    return sites_list


def move_copy_core(src, dest, copy_only=False, force=False):
    """
    Moves or copies a file from the source to the destination.

    This function attempts to copy or move a file from the source path to the destination path.
    If the operation is successful, it logs the action and returns the path of the moved/copied file.
    If the operation fails, it logs the error and returns None.

    Parameters
    ----------
    src : str
        The source file path.
    dest : str
        The destination file path.
    copy_only : bool, optional
        If True, the file is copied instead of moved. Default is False.
    force : bool, optional
        Force the move/copy if the file already exists
        Default is False

    Returns
    -------
    str or None
        The path of the moved/copied file if the operation is successful, None otherwise.
    """
    mvcp = "copied" if copy_only else "moved"

    if os.path.isfile(dest) and not force:
        logger.info(f"{dest} exists and kept (force={force})")
        return dest

    try:
        # we prefer a copy rather than a move, mv can lead to some error
        file_moved = shutil.copy2(src, dest)
        file_moved = os.path.abspath(file_moved)
        # file_moved = shutil.move(src, dest)
        logger.debug("file " + mvcp + " to final destination: %s", file_moved)
    except Exception as e:
        logger.error("Error for: %s", src)
        logger.error("Exception raised: %s", e)
        file_moved = None

    if file_moved and (not copy_only) and os.path.isfile(src):
        os.remove(src)

    return file_moved


def log_tester():
    """
    Tests if the logger is working.

    This function checks if the logger is working by logging a test message.
    If the logger is not working, it raises an exception.

    Raises
    ------
    Exception
        If the logger is not working.
    """
    from logging_tree import printout
    printout()

    logger.debug("level debug")
    logger.info("level info")
    logger.warning("level warning")
    logger.error("level error")
    logger.critical("level critical")

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 18/09/2024 18:24:43

@author: psakic
"""

import os
import pandas as pd
import autorino.convert as arocnv
import autorino.common as arocmn
import logging
import autorino.cfgenv.env_read as aroenv
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed

import multiprocessing as mp
import geodezyx.utils

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


def convert_rnx(
    inp_raws,
    out_dir,
    out_structure="%Y/%j",
    tmp_dir=None,
    log_dir=None,
    rinexmod_options=None,
    metadata=None,
    force_rnx=False,
    force_raw=False,
    raw_out_dir=None,
    raw_out_structure=None,
    processes=1,
    filter_prev_tables=False,
    converter="auto"
):
    """
    Frontend function that performs RAW > RINEX conversion.

    Parameters
    ----------
    inp_raws : list
        The input RAW files to be converted.
        The input can be:
        * a python list
        * a text file path containing a list of files
        * a tuple containing several text files path
        * a directory path.
    out_dir : str
        The output directory where the converted files will be stored.
    out_structure : str, optional
        The structure of the output directory.
        If provided, the converted files will be stored in a subdirectory of out_dir following this structure.
        See README.md for more information.
        Typical values are '<SITE_ID4>/%Y/' or '%Y/%j/'.
        Default value is '%Y/%j/'.
    tmp_dir : str, optional
        The temporary directory used during the conversion process.
        If not provided, it defaults to <out_dir>/tmp_convert_rnx.
        Defaults to None.
    log_dir : str, optional
        The directory where logs will be stored. If not provided, it defaults to tmp_dir.
         Defaults to None.
    rinexmod_options : dict, optional
        The options for modifying the RINEX files during the conversion. Defaults to None.
    metadata : str or list, optional
        The metadata to be included in the converted RINEX files.
        Possible inputs are:
         * list of string (sitelog file paths),
         * single string (single sitelog file path)
         * single string (directory containing the sitelogs)
         * list of MetaData objects
         * single MetaData object.
         Defaults to None.
    force_rnx : bool, optional
        If set to True, the conversion will be forced even if the output files already exist.
        Defaults to False.
    force_raw : bool, optional
        If set to True, the RAW file archiving will be forced even if the output files already exist.
        Defaults to False.
    raw_out_dir : str, optional
        Directory where RAW files will be archived.
        No move/delete will occur, your input RAW files are sacred.
        Defaults to None.
    raw_out_structure : str, optional
        Structure for archiving RAW files.
        Defaults to `out_structure` if not provided.
    processes : int, optional
        Number of processes to use for parallel conversion. Default is 1.
    filter_prev_tables : bool, optional
        If True, filters and skip previously converted files
        with tables stored in the tmp tables directory.
        Default is False.
    converter: str, optional
        The software to be used for converting the RAW files to RINEX.
        Default is "auto".

    Returns
    -------
    None
    """

    if not metadata:
        logger.warning("No metadata (sitelogs...) provided while highly recommended.")
        logger.warning("Be sure of what you are doing! (check -m option)")

    tmp_dir = tmp_dir or os.path.join(out_dir, "tmp_convert_rnx")
    log_dir = log_dir or tmp_dir
    out_dir_use = os.path.join(out_dir, out_structure) if out_structure else out_dir

    # get the raw files as a list depending on the input type
    inp_raws_lis = arocmn.import_files(inp_raws)

    ###### Convert RAW > RINEX files
    inp_raws_lis_cnk = geodezyx.utils.chunkIt(inp_raws_lis, processes)

    args_wrap = []
    for raws in inp_raws_lis_cnk:
        args = (
            raws,
            out_dir_use,
            tmp_dir,
            log_dir,
            metadata,
            force_rnx,
            rinexmod_options,
            filter_prev_tables,
            converter,
        )
        args_wrap.append(args)

    # Parallel RAW > RINEX conversion

    #### ++++ new style concurrent.futures
    cnv_out_lis = []
    # pool_exec = ThreadPoolExecutor
    pool_exec = ProcessPoolExecutor
    with pool_exec(max_workers=processes) as executor:
        futures = [executor.submit(convert_raw_wrap, args) for args in args_wrap]
        for f in as_completed(futures):
            cnv_out_lis.append(f.result())

    #### ++++ classic multiprocessing
    # pool = mp.Pool(processes=processes)
    # try:
    #     _ = pool.map(convert_raw_wrap, args_wrap, chunksize=1)
    # except Exception as e:
    #     logger.error("error in the pool.map : %s", e)
    #
    # pool.close()

    ###### Archive the RAW files
    if raw_out_dir:
        logger.info(">>>>>> RAW files archive")
        if not raw_out_structure:
            raw_out_structure = out_structure
        raw_out_dir_use = str(os.path.join(raw_out_dir, raw_out_structure))

        cnv_table_cat = pd.concat([cnv.table for cnv in cnv_out_lis], ignore_index=True)

        cpy_raw = arocmn.StepGnss(raw_out_dir_use, tmp_dir, log_dir, metadata=metadata)
        debug_print = False

        # the table from the ConvertGnss object
        # is necessary to get the epoch
        cpy_raw.load_tab_prev_tab(cnv_table_cat, new_inp_is_prev="inp")
        cpy_raw.print_table() if debug_print else None
        cpy_raw.filter_na(["epoch_srt", "epoch_end"])
        cpy_raw.print_table() if debug_print else None
        cpy_raw.guess_out_files()
        cpy_raw.print_table() if debug_print else None
        cpy_raw.filter_ok_out()
        cpy_raw.print_table() if debug_print else None
        cpy_raw.move_files(mode="inpout", force=force_raw, copy_only=True)
        cpy_raw.print_table() if debug_print else None

    return cnv_out_lis


def convert_raw_wrap(args):
    """
    Wrapper function for converting RAW GNSS files to RINEX format.

    Parameters
    ----------
    args : tuple
        A tuple containing the following arguments:
        - raws : list
            List of RAW files to be converted.
        - out_dir_use : str
            The output directory where the converted files will be stored.
        - tmp_dir : str
            Temporary directory used during the conversion process.
        - log_dir : str
            Directory where logs will be stored.
        - metadata : str or list
            Metadata to be included in the converted RINEX files.
        - force_rnx : bool
            If True, forces the conversion even if output files already exist.
        - rinexmod_options : dict
            Options for modifying the RINEX files during the conversion.
        - filter_prev_tables : bool
            If True, filters and skips previously converted files with
            tables stored in the tmp tables directory.
        - converter : str
            The software to be used for converting the RAW files to RINEX.

    Returns
    -------
    arocnv.ConvertGnss
        An instance of the `ConvertGnss` class containing the results of the conversion.
    """
    (
        raws,
        out_dir_use,
        tmp_dir,
        log_dir,
        metadata,
        force_rnx,
        rinexmod_options,
        filter_prev_tables,
        converter
    ) = args
    # Initialize the ConvertGnss object with the provided directories and metadata
    cnv = arocnv.ConvertGnss(out_dir_use, tmp_dir, log_dir, metadata=metadata)
    # Load the list of RAW files to be converted
    cnv.load_tab_filelist(raws)
    # Perform the conversion with the specified options
    cnv.convert(
        force=force_rnx,
        rinexmod_options=rinexmod_options,
        filter_prev_tables=filter_prev_tables,
        converter=converter,
    )
    # Return the ConvertGnss object containing the conversion results
    return cnv



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/01/2025 09:50:23

@author: psakic
"""
import matplotlib.pyplot as plt

import autorino.cfgenv as aroenv
import autorino.check as arochk
import autorino.common as arocmn
import pandas as pd
import os
import logging
from geodezyx import utils

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


def check_rnx(
    inp_dir_parent,
    inp_dir_structure,
    epoch_start,
    epoch_end,
    sites_list=[],
    output_dir=None,
):
    """
    Checks the presence of RINEX files in the input directory over a specified time range,
    and computes their completeness ratio.

    Parameters
    ----------
    inp_dir_parent : str
        The parent input directory.
    inp_dir_structure : str
        The input directory structure.
    epoch_start : str
        The start epoch.
    epoch_end : str
        The end epoch.
    sites_list : list, optional
        A list of site identifiers to filter the check.
    output_dir : str, optional
        The output directory.

    Returns
    -------
    tabu_chk_col : str
        The tabulated string of the check table with colored values.
    tabu_chk_bnw : str
        The tabulated string of the check table with black and white values.
    df_chk_sum  : pd.DataFrame
        The values of the check table summarized in a dataframe.
    df_chk_full_stats : pd.DataFrame
        The full statistics of the check table.
    """

    inp_dir = os.path.join(inp_dir_parent, inp_dir_structure)
    eporng = arocmn.EpochRange(epoch_start, epoch_end)

    if sites_list:
        sites_use = sites_list
    else:
        sites_use = arocmn.guess_sites_list(inp_dir_parent)

    chk_tab_stk = []
    chk_tab_stats_stk = []
    for site in sites_use:
        chk = arochk.CheckGnss(
            inp_dir=str(inp_dir),
            site={"site_id": site},
            epoch_range=eporng
        )
        chk.check()
        chk_tab_stk.append(chk.table)
        chk_tab_stats_stk.append(chk.table_stats)

    df_chk_table_cat = pd.concat(chk_tab_stk)
    df_chk_full_stats = pd.concat(chk_tab_stats_stk)

    tabu_chk_col, tabu_chk_bnw, df_chk_sum = arochk.get_tabult_raw(
        df_chk_table_cat, short_label=True
    )

    logger.info("Check: \n" + tabu_chk_col)

    if output_dir:
        checkrnx_output(
            output_dir,
            eporng,
            tabu_chk_col,
            tabu_chk_bnw,
            df_chk_sum,
            df_chk_full_stats,
        )

    return tabu_chk_col, tabu_chk_bnw, df_chk_sum, df_chk_full_stats


# defcheckrnx_analyz
# checkrnx_format
def checkrnx_output(
    output_dir, eporng, tabu_chk_col, tabu_chk_bnw, df_chk_sum, df_chk_full_stats
):
    """
    This function saves the check_rnx results in the output directory.

    Parameters
    ----------
    output_dir : str
        The output directory.
    eporng : EpochRange
        The epoch range.
    tabu_chk_col : str
        The tabulated string of the check table with colored values.
    tabu_chk_bnw : str
        The tabulated string of the check table with black and white values.
    df_chk_sum  : pd.DataFrame
        The values of the check table summarized in a dataframe.
    df_chk_full_stats : pd.DataFrame
        The full statistics of the check table.

    Returns
    -------
    None
    """

    prefix = "_".join(
        (
            utils.get_timestamp(),
            eporng.epoch_start.strftime("%Y-%j"),
            eporng.epoch_end.strftime("%Y-%j"),
        )
    )

    output_dir_use = utils.create_dir(os.path.join(output_dir, prefix))
    ### csv
    summ_csv = os.path.join(output_dir_use, prefix + "_check_rnx_summ.csv")
    full_csv = os.path.join(output_dir_use, prefix + "_check_rnx_full.csv")
    df_chk_sum.to_csv(summ_csv)
    df_chk_full_stats.to_csv(full_csv)

    ### plot
    df_chk_sum.plot()
    fig = plt.gcf()
    utils.figure_saver(
        fig, output_dir_use, prefix + "_check_rnx_plot", outtype=(".png", ".pdf")
    )
    ### pretty print tabulate
    tabu_col_txt = os.path.join(output_dir_use, prefix + "_check_rnx_tabu_col.txt")
    tabu_bnw_txt = os.path.join(output_dir_use, prefix + "_check_rnx_tabu_bnw.txt")
    utils.write_in_file(tabu_chk_col, tabu_col_txt)
    utils.write_in_file(tabu_chk_bnw, tabu_bnw_txt)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/01/2025 18:53:22

@author: psakic
"""

import termcolor
import pandas as pd
import tabulate
import os


def color(val):
    """
    This function returns a color according to the value of the input.

    Parameters
    ----------
    val: float
        The value to be colored.

    Returns
    -------
    str: color
    """
    if val > 96.0:
        return "cyan"
    elif val <= 1.0:
        return "magenta"
    elif 50. >= val > 1.0:
        return "red"
    else:
        return "yellow"


def colorize_list(list_inp):
    """
    This function colorizes a list of values.

    Parameters
    ----------
    list_inp : list
        The list of values to be colorized.

    Returns
    -------
    list: list of colorized values

    """
    if not 'TERM' in os.environ:
        os.environ['TERM'] = 'xterm-256color'  # Set terminal type to xterm-256color if not set

    return [termcolor.colored(str(e), color(e)) for e in list_inp]


def get_tabult_raw(chk_tab, short_label=False):
    """
    This function returns a tabulated string of the check table.

    Parameters
    ----------
    chk_tab : pd.DataFrame
        The check table.

    short_label : bool, optional

    Returns
    -------
    tabu_chk_col: str
        The tabulated string of the check table with colored values.
    tabu_chk_bnw: str
        The tabulated string of the check table with black and white values.
    df_chk_sum: pd.DataFrame
        The values of the check table summarized in a dataframe.

    """
    chk_tab = chk_tab.sort_values(["epoch_srt", "site"])
    sites = list(chk_tab["site"].unique())

    if short_label:
        print(sites)
        sites = [s[:4] for s in sites]
        fmt_time = "%Y-%j"
    else:
        fmt_time = "%Y-%j %H:%M"

    tab_chk_raw_col = [["epoch_srt"] + sites]
    tab_chk_raw_bnw = [["epoch_srt"] + sites]
    flt_data_stk = []

    for epo, chk_epo in reversed(list(chk_tab.groupby("epoch_srt"))):
        epo = pd.Timestamp(epo)
        l_flt = [epo] + list(chk_epo["%"].tolist())
        l_str_col = [epo.strftime(fmt_time)] + list(colorize_list(chk_epo["%"].tolist()))
        l_str_bnw = [epo.strftime(fmt_time)] + chk_epo["%"].tolist()
        flt_data_stk.append(l_flt)
        tab_chk_raw_col.append(l_str_col)
        tab_chk_raw_bnw.append(l_str_bnw)


    tabu_chk_col = tabulate.tabulate(tab_chk_raw_col,
                                     headers="firstrow",
                                     tablefmt="grid")
    tabu_chk_bnw = tabulate.tabulate(tab_chk_raw_bnw,
                                    headers="firstrow",
                                    tablefmt="grid")

    df_chk_sum = pd.DataFrame(flt_data_stk, columns=tab_chk_raw_col[0])
    df_chk_sum.set_index("epoch_srt", inplace=True)

    return tabu_chk_col, tabu_chk_bnw, df_chk_sum



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  8 15:47:58 2024

@author: psakic
"""
import logging
import re

import dateparser
import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
import datetime as dt

import autorino.common as arocmn
import autorino.cfgenv.env_read as aroenv

logger = logging.getLogger("autorino")
logger.setLevel(aroenv.ARO_ENV_DIC["general"]["log_level"])


def epoch_range_intrpt(epo_inp):
    """
    This function interprets an input to get an output EpochRange object. The input can either be a tuple,
    typically in the form of (epo1, epo2, period), or an instance of the EpochRange class. If the input is
    an EpochRange object, it is returned as is. If the input is a tuple, a new EpochRange object is created
    using the elements of the tuple.

    Parameters
    ----------
    epo_inp : tuple or EpochRange
        The input to be interpreted. If it's a tuple, it should be in the form of (epo1, epo2, period).

    Returns
    -------
    epo_range_out : EpochRange
        The interpreted EpochRange object.
    """
    if type(epo_inp) is arocmn.EpochRange:
        epo_range_out = epo_inp
    elif type(epo_inp) is tuple and len(epo_inp) == 3:
        epo_range_out = arocmn.EpochRange(*epo_inp)
    elif type(epo_inp) is tuple and len(epo_inp) == 2:
        epo_range_out = arocmn.EpochRange(epo_inp[0], epo_inp[1])
    else:
        logger.error("epoch range input not understood")
        raise Exception

    return epo_range_out


def datepars_intrpt(date_inp, tz=None, tz_if_naive="UTC"):
    """
    This function interprets a string or datetime-like object to a Pandas Timestamp.
    It also applies a timezone (UTC by default). Note that rounding does not take place here
    as rounding is not a parsing operation.

    Parameters
    ----------
    date_inp : str or datetime-like
        The input date to be interpreted.
    tz : str, optional
        The timezone to be applied.
        The default is None.
    tz_if_naive : str, optional
        The timezone to be applied if the input date is timezone-naive.
        The default is "UTC".

    Returns
    -------
    date_out : datetime
        The interpreted date as a python native datetime.

    Note
    ----
    If the input date is a string, it is parsed using the dateparser library.
    If the input date is a datetime-like object, it is converted first to a Pandas Timestamp.
    If the resulting date is a NaT (Not a Time) type, it is returned as is.
    If the resulting date does not have a timezone, the specified timezone is applied.
    """
    ### INTERPRET THE INPUT
    if not isinstance(date_inp, str):
        date_out = pd.Timestamp(date_inp)
    ## date_inp is a str
    else:
        ### Must handle the case of day of year separately
        doy_pattern_1 = r"^\d{4}-\d{1,3}$"
        doy_pattern_2 = r"^\d{4}/\d{1.3}$"
        # YYYY-DDD
        if re.match(doy_pattern_1, date_inp):
            date_out = pd.Timestamp(dt.datetime.strptime(date_inp, "%Y-%j"))
        # YYYY/DDD
        elif re.match(doy_pattern_2, date_inp):
            date_out = pd.Timestamp(dt.datetime.strptime(date_inp, "%Y/%j"))
        ### regular case
        else:
            date_out = pd.Timestamp(dateparser.parse(date_inp).isoformat())
            # .isoformat() is to correctly handle the time zone,
            # without weird pytz's objects used by dateparser
            # like <StaticTzInfo 'UTC\+00:00'>

    ### MANAGE THE TIMEZONE
    ### NaT case. can not support tz
    if isinstance(date_out, pd._libs.tslibs.nattype.NaTType):
        return date_out
    ### if the date is timezone-naive, apply the tz_if_naive
    if not date_out.tz:
        logger.warning("date %s is timezone-naive. Applying tz %s", date_out, tz_if_naive)
        date_out = pd.Timestamp(date_out, tz=tz_if_naive)
    ### apply the tz
    if tz:
        date_out = date_out.tz_convert(tz)

    ### OUTPUT A NATIVE DATETIME
    date_out = date_out.to_pydatetime()

    return date_out


def dates_list2epoch_range(dates_list_inp, period=None, round_method="floor"):
    """
    Converts a list of dates to an EpochRange.

    Parameters
    ----------
    dates_list_inp : iterable
        The input list of dates.
    period : str, optional
        The rounding period. If not provided, the period is determined as the unique difference
         between consecutive dates in the input list. The default is None.
    round_method : str, optional
        The method used for rounding the epochs. The default is 'floor'.

    Returns
    -------
    EpochRange
        The converted EpochRange.

    Note
    ----
    The current method for determining the period when not provided is a simple calculation of
    the unique difference between consecutive dates.
    This may not be the most accurate or desired method and should be improved in future iterations of this function.
    """
    epoch1 = np.min(dates_list_inp)
    epoch2 = np.max(dates_list_inp)

    if period:
        period_use = period
    else:
        period_use = np.unique(np.diff(dates_list_inp))[0]  # poor, must be improved

    epo_out = arocmn.EpochRange(
        epoch1, epoch2, period=period_use, round_method=round_method
    )

    return epo_out


def round_date_legacy(date_in, period, round_method="floor"):
    """
    low-level function to round a Pandas Serie or a datetime-like object
    according to the "ceil", "floor", "round", "none" approach

    Parameters
    ----------
    date_in : Pandas Serie or a datetime-like object
        Input date .
    period : str, optional
        the rounding period.
        Use the pandas' frequency aliases convention (see bellow for details).
    round_method : str, optional
        round method: 'ceil', 'floor', 'round', 'none'. The default is "floor".

    Returns
    -------
    date_out : Pandas Serie or datetime-like object (same as input)
        rounded date.

    Note
    ----
    Pandas' frequency aliases memo
    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases

    """

    # we separate the Series case of the simple datetime-like
    # both for ease and performance reason

    if pd.isna(date_in):  ### NaT case
        date_out = date_in
    elif isinstance(date_in, pd.Series):
        date_use = date_in

        if round_method == "ceil":
            date_out = date_use.dt.ceil(period)
        elif round_method == "floor":
            date_out = date_use.dt.floor(period)
        elif round_method == "round":
            date_out = date_use.dt.round(period)
        elif round_method == "none":
            date_out = date_use
        else:
            logger.critical("round_method not understood")
            raise Exception
    else:
        date_typ = type(date_in)
        if date_typ in (pd.Timedelta,):
            date_use = pd.Timedelta(date_in)
        else:
            date_use = pd.Timestamp(date_in)

        if round_method == "ceil":
            date_out = date_use.ceil(period)
        elif round_method == "floor":
            date_out = date_use.floor(period)
        elif round_method == "round":
            date_out = date_use.round(period)
        elif round_method == "none":
            date_out = date_use
        else:
            logger.critical("round_method not understood")
            raise Exception

        #++++ back to the original type
        if date_typ in (dt.datetime,):
            date_out = date_out.to_pydatetime()
        else:
            date_out = date_typ(date_out)

    return date_out


def round_date(date_in, period, round_method="floor"):
    """
    low-level function to round a Pandas Serie or a datetime-like object
    according to the "ceil", "floor", "round", "none" approach

    Parameters
    ----------
    date_in : Pandas Serie or a datetime-like object
        Input date .
    period : str, optional
        the rounding period.
        Use the pandas' frequency aliases convention (see bellow for details).
    round_method : str, optional
        round method: 'ceil', 'floor', 'round', 'none'. The default is "floor".

    Returns
    -------
    date_out : Pandas Serie or datetime-like object (same as input)
        rounded date.

    Note
    ----
    Pandas' frequency aliases memo
    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases

    """

    # ++++ NaT case
    if pd.isna(date_in):
        return date_in

    # ++++ Series case
    if isinstance(date_in, pd.Series):
        return getattr(date_in.dt, round_method)(period)

    # ++++ Singleton case
    date_use = pd.Timedelta(date_in) if isinstance(date_in, pd.Timedelta) else pd.Timestamp(date_in)
    date_out = getattr(date_use, round_method)(period) if round_method != "none" else date_use

    # ++++ back to the original type
    date_out = date_out.to_pydatetime() if isinstance(date_in, dt.datetime) else type(date_in)(date_out)

    return date_out


def round_epochs(
    epochs_inp, period="1d", rolling_period=False, rolling_ref=-1, round_method="floor"
):
    """
    High-level function to round several epochs to a common one.
    Useful to group and then splice RINEX

    Use it with pandas .groupby in a further step

    Parameters
    ----------
    epochs_inp : iterable
        The input epochs expected to be rounded.
    period : str, optional
        the rounding period.
        Use the pandas' frequency aliases convention (see bellow for details).
        The default is '1d'.
    rolling_period : bool, optional
        Whether to use a rolling period for splicing the RINEX files.
        If False, the spliced files will be based only on the "full" period provided,
        i.e. Day1 00h-24h, Day2 00h-24h, etc.
        If True, the spliced files will be based on the rolling period.
        i.e. Day1 00h-Day2 00h, Day1 01h-Day2 01h, Day1 02h-Day2 02h etc.
        Defaults to False.
    rolling_ref :  datetime-like or int, optional
        The reference for the rolling period.
        If datetime-like object, use this epoch as reference.
        If integer, use the epoch of the corresponding index
        Use -1 for the last epoch for instance.
        The default is -1.
    round_method : str, optional
        round method: 'ceil', 'floor', 'round'. The default is "floor".

    Returns
    -------
    epochs_rnd :
        Pandas Series of Timestamp or Timedelta

    Note
    ----
    Pandas' frequency aliases memo
    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases

    """

    epochs_use = pd.Series(epochs_inp)

    if not rolling_period:
        epochs_rnd = arocmn.round_date(epochs_use, period, round_method)
    else:
        if type(rolling_ref) is int:
            rolling_ref_use = epochs_use.iloc[rolling_ref]
        else:
            rolling_ref_use = rolling_ref

        ### add one second to be sure that the rolling_ref is included in a group
        rolling_ref_use = rolling_ref_use + np.timedelta64(1, "s")

        roll_diff = epochs_use - rolling_ref_use

        epochs_rnd = (
            arocmn.round_date(roll_diff, period, round_method) + rolling_ref_use
        )

    return epochs_rnd


def timedelta2freq_alias(timedelta_in):
    """
    Time representation conversion

    Timedelta (from datetime, numpy or pandas) => Pandas' Frequency alias

    Parameters
    ----------
    timedelta_in : timedelta-like
        POSIX Time.  Can *NOT YET* handle several timedelta in a list.

    Returns
    -------
    offset : string
        Converted pandas' frequency alias

    Note
    ----
    Pandas' frequency aliases memo
    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases
    """

    offset = to_offset(pd.Timedelta(timedelta_in))
    return offset.freqstr


def create_dummy_epochrange():
    """
    Create a fake/dummy EpochRange object
    for test/development purpose

    Returns
    -------
    ses : EpochRange object
        dummy EpochRange object.

    """

    epo = arocmn.EpochRange(epoch1=pd.NaT, epoch2=pd.NaT, period="15min")
    return epo


def iso_zulu_epoch(epo_in):
    """
    Convert an input epoch to ISO 8601 format with Zulu time (UTC).

    Parameters
    ----------
    epo_in : datetime-like
        The input epoch to be converted.

    Returns
    -------
    str
        The epoch in ISO 8601 format with Zulu time (UTC).
    """
    return pd.Timestamp(epo_in).isoformat().replace("+00:00", "Z")
